#  Minimal code for complete regular temperament searches.
#  But not self-sufficient
#  Supplement to http://x31eq.com/complete.pdf

from __future__ import division

#
#  definitions
#

import math

# Intervals are measured in octaves.
# In general, the first entry in a list of primes is taken to be
# the equivalence interval.
# Tenney weighting is implied.
primeNumbers = (2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31)
primes = [math.log(f)/math.log(2) for f in primeNumbers]

#
#  Equal Temperaments
#

def getEqualTemperaments(primes, kmax, Emax, dmax=None):
    if dmax is None:
        dmax = int(math.sqrt(kmax/Emax))
    # Check we get all period divisions
    dmax = max(dmax, largestOctaveDivision(primes, kmax))
    k2od2 = (kmax/(dmax+1))**2
    for nNotes in xrange(1, dmax+1):
        cutoff = k2od2/nNotes**2 + Emax**2
        for mapping in varLimitedMappings(nNotes, cutoff, primes):
            yield mapping

def varLimitedMappings(nNotes, maxVariance, primes):
    """Return all ET mappings with the given number of notes and the
    optimal weighted variance less than that given.
    """
    # convert the maximum variance into something that
    # will steadily increase as the mapping increases
    limit = maxVariance*len(primes) # correction for primes[0]?

    def moreLimitedMappings(mapping, tot, tot2):
        """mapping - the ET mapping with a new entry
        tot - running total of w
        tot2 - running total of w squared
        """
        i = len(mapping)
        weightedSize = mapping[-1]/primes[i-1]/nNotes
        tot += weightedSize
        tot2 += weightedSize**2
        error = tot2 - tot**2/i
        assert error <= limit # we shouldn't be here otherwise
        if i==len(primes):
            # recursion stops here
            yield mapping
        else:
            target = primes[i]*nNotes
            deficit = math.sqrt((i+1)*(limit - error)/i)
            nmin = int(math.ceil(target*(tot/i - deficit)))
            nmax = int(math.floor(target*(tot/i + deficit)))
            for guess in range(nmin, nmax+1):
                results = moreLimitedMappings(mapping+[guess], tot, tot2)
                for result in results:
                    yield result

    return moreLimitedMappings([nNotes], 0.0, 0.0)

#
#  Linear temperaments constructed from equal temperaments or mappings
#

# Some things can't be simplified much relative to regular.py
from regular import allMappings, STDComplexityFinder, VarianceErrorFinder

def survey(primes, kmax, Emax):
    ets = getEqualTemperaments(primes, kmax, Emax)
    return list(getLinearTemperaments(primes, ets, kmax, Emax))

def getLinearTemperaments(primes, ets, kmax, Emax):
    oldInvariants = set()
    for et in ets:
        mappings = allMappings(primes, et, kmax, Emax,
                STDComplexityFinder, VarianceErrorFinder)
        for gen, permap, genmap in mappings:
            key = invariant(permap, genmap)
            if key not in oldInvariants:
                oldInvariants.add(key)
                yield permap, genmap

def largestOctaveDivision(primes, kmax):
    """Returns the largest number of periods per octave
    for a rank 2 temperament with complexity within kmax
    """
    np = len(primes)
    return int(kmax*np*max(primes) / primes[0]/math.sqrt(np-1))

def invariant(permap, genmap):
    """Return something that uniquely identifies
    the temperament class
    """
    if filter(None, genmap)[0] < 0:
        multiplier = -permap[0]
    else:
        multiplier = permap[0]
    return tuple([multiplier*m for m in genmap])

#
#  Test code
#
def test():
    """Duplicate the examples from complete.pdf"""
    limit7, limit11, limit13 = primes[:4], primes[:5], primes[:6]
    assert list(getEqualTemperaments(limit11, 2.788, 0.484/1200))==[
            [72, 114, 167, 202, 249]]
    assert list(getEqualTemperaments(limit11, 2.788, 0.484/1200, dmax=71))==[
            [31, 49, 72, 87, 107]]
    assert largestOctaveDivision(limit11, 2.788)==24
    assert survey(limit11, 2.788, 0.484/1200)==[]
    assert survey(limit11, 2.7881, 0.484/1200)==[]
    assert list(getEqualTemperaments(limit11, 2.7881, 0.4841/1200))==[
            [72, 114, 167, 202, 249]]
    assert list(getEqualTemperaments(limit11, 2.7881,0.4841/1200,dmax=71))==[
            [31, 49, 72, 87, 107]]
    assert largestOctaveDivision(limit11, 2.7881)==24
    assert survey(limit11, 2.788, 0.4841/1200)==[]
    miracle = ((1,1,3,3,2), (0,6,-7,-2,15))
    assert survey(limit11, 2.7881, 0.4841/1200)==[miracle]
    assert survey(limit11, 2.788, 1/1200)==[
            ((1,-1,0,1,-3), (0,10,9,7,25))]
    assert survey(limit11, 2.7881, 0.851/1200)==[miracle]
    assert survey(limit11, 4.064, 0.484/1200)==[
                ((2,1,5,2,8), (0,6,-1,10,-3))]
    assert survey(limit11, 4.063, 0.4841/1200)==[miracle]
    assert len(survey(limit11, 4.063, 0.851/1200))==7
    assert survey(limit11, 3.75, 0.65/1200)==[miracle]
    nearMiracle = survey(limit11, 3.78, 0.68/1200)
    assert len(nearMiracle)==3
    assert miracle in nearMiracle
    assert ((1,1,-1,3,6),(0,3,17,-1,-13)) in nearMiracle
    assert ((1,-5,0,-3,-7),(0,17,6,15,27)) in nearMiracle

    assert survey(limit7, 4.724, 0.03/1200)==[
            ((9,15,22,26),(0,-2,-3,-2))]
    assert survey(limit7, 4.724, 0.22/1200)==[
            ((9,15,22,26),(0,-2,-3,-2))]
    assert survey(limit7, 4.724, 0.029/1200)==[]

    mystery = ((29,46,68,82,101,108),(0,0,-1,-1,-1,-1))
    miracle13 = ((1,1,3,3,2,7),(0,6,-7,-2,15,-34))
    miraclish = ((2,4,7,7,9,11),(0,-6,-17,-10,-15,-26))
    mysterious = survey(limit13, 4.837, 0.513/1200)
    assert mystery in mysterious
    assert miracle13 in mysterious
    assert len(mysterious)==2
    assert survey(limit13, 4.836, 0.513/1200)==[miracle13]
    mysterious = survey(limit13, 4.9, 0.52/1200)
    assert mystery in mysterious
    assert miracle13 in mysterious
    assert miraclish in mysterious
    assert len(mysterious)==3

    wideets = getEqualTemperaments(limit11, 4, 0.5)
    for i in range(100):
        assert i < 11
        if wideets.next() != [1,0,-1]:
            break
    # refresh generator
    wideets = getEqualTemperaments(limit11, 4, 0.5)

if __name__=="__main__":
    test()
