"""Code for converting between mappings and unison vectors"""

from __future__ import division

import regutils, matfunc

def more_uvs(mapping, plimit):
    """Returns a selection of unison vectors"""
    uvs = unison_vectors(mapping, plimit)
    if len(uvs)==1:
        return uvs
    others = secondary_uvs(uvs, plimit)
    unison = (0,)*len((plimit))
    uvset = set(map(tuple, others))
    uvset.add(unison)
    for uva in uvs:
        for uvb in others:
            zipped = zip(uva, uvb)
            newuvs = [[x+y for (x, y) in zipped], [x-y for (x, y) in zipped]]
            normalize_intervals(newuvs, plimit)
            for newuv in newuvs:
                if tuple(newuv) not in uvset:
                    uvset.add(tuple(newuv))
    uvset.remove(unison)
    return tenney_sort(uvset, plimit)[:len(uvs)*2]

def unison_vectors(mapping, plimit):
    """Returns a basis of unison vectors"""
    basis = saturated_basis(mapping)
    reduced = TLLL(basis, plimit)
    normalize_intervals(reduced, plimit)
    return reduced

def temper_out(uvs):
    basis = saturated_basis(uvs)
    return regutils.lattice_reduction(basis)

def saturated_basis(vectors):
    """Kernel basis without contorsion"""
    return saturate(kernel_basis(vectors))

#
# Code to simplify results
#

def secondary_uvs(uvs, plimit):
    """Single permutation of the basis"""
    assert uvs
    if len(uvs)==1:
        return uvs
    result = [uv for uv in uvs]
    for i, uv1 in enumerate(uvs):
        for uv2 in uvs[i+1:]:
            z = zip(uv1, uv2)
            result += [[x+y for (x, y) in z], [x-y for (x, y) in z]]
    return tenney_sort(uvs, plimit)

def tenney_sort(vectors, plimit):
    """Sort intervals by Tenney harmonic distance,
    and make positive
    """
    decorated = [(thd(uv, plimit), uv) for uv in vectors]
    decorated.sort()
    result = [uv for (x, uv) in decorated]
    normalize_intervals(result, plimit)
    return result

def normalize_intervals(vectors, plimit):
    """Make all intervals positive -- destructive"""
    for i, row in enumerate(vectors):
        size = sum([v*p for (v,p) in zip(row, plimit)])
        if size < 0.0:
            vectors[i] = [-x for x in row]

def reverse_hermite(vectors):
    rev = [reversed(v) for v in vectors]
    hermite = regutils.lattice_reduction(rev)
    return [list(reversed(v)) for v in hermite]

def thd(vector, plimit):
    """Tenney Harmonic Distance"""
    return sum([abs(v*p) for (v,p) in zip(vector, plimit)])

def TLLL(vectors, limit):
    """Tenney weighted LLL reduction of intervals.
    Use reciprocal weighting for intervals.
    """
    weights = [p*p for p in limit]
    def tenneyEuclid(u, v):
        assert len(u)==len(v) and len(u) <= len(weights)
        return sum([x*y*w for x, y, w in zip(u, v, weights)])
    return regutils.LLL(vectors, tenneyEuclid)

#
# Code to find a kernel
#

def kernel_basis(vectors):
    """Integer basis that may have contorsion"""
    # Well known algorithm.
    # c.f. http://en.wikipedia.org/wiki/Null_space#Basis
    # Should be fine for integers.
    n_cols = len(vectors[0])
    if __debug__:
        for vector in vectors:
            assert len(vector)==n_cols
    reduced = ref(vectors)
    del reduced[regutils.rank(reduced):]
    pivots, pivot_product = [], 1
    for row in reduced:
        for i, square in enumerate(row):
            if square != 0:
                pivots.append(i)
                pivot_product *= square
                break
        assert True, "Pivot not found"
    non_pivots = [i for i in range(n_cols) if i not in pivots]
    rows = [None]*n_cols
    assert n_cols == len(pivots) + len(non_pivots)
    for i, j in enumerate(pivots):
        multiplier = -pivot_product//reduced[i][j]
        rows[j] = [multiplier * reduced[i][k] for k in non_pivots]
    for i, j in enumerate(non_pivots):
        row = [0]*len(non_pivots)
        row[i] = pivot_product
        rows[j] = row
    result = zip(*rows)
    if __debug__:
        for row in rows:
            assert row is not None
        # check the null space is correct
        for row in result:
            assert regutils.tempersOut(vector, row)
    return result

def ref(vectors):
    """Reduced row echelon form with integers"""
    reduced = regutils.lattice_reduction(vectors)
    for row in range(1, len(reduced)):
        j = 1
        while reduced[row][j] == 0:
            j += 1
            if j == len(reduced[row]):
                return reduced
        bottom_square = reduced[row][j]
        for i in range(row):
            other_square = reduced[i][j]
            if other_square != 0:
                reduced[i] = [x*bottom_square - y*other_square
                        for (x, y) in zip(reduced[i], reduced[row])]
            assert reduced[i][j]==0
    return reduced

#
# Code to deal with (con)torsion
#

def saturate(vectors):
    # c.f. http://www.wstein.org/papers/hnf/
    # pernet-stein-fast_computation_of_hnf_of_random_integer_matrices.pdf
    assert vectors
    hermite = regutils.lattice_reduction(vectors)
    transpose = zip(*hermite)
    double_hermite = regutils.lattice_reduction(transpose)
    if len(vectors)==1:
        gcd = double_hermite[0][0]
        return [[x//gcd for x in vectors[0]]]
    transformation = matfunc.Square(double_hermite[:len(vectors)]).inverse()
    saturated = transformation.tr().mmul(matfunc.Matrix(hermite))
    return integer_matrix(saturated)

def contorsion(mapping):
    hermite = regutils.lattice_reduction(mapping)
    transpose = zip(*hermite)
    double_hermite = regutils.lattice_reduction(transpose)
    total = 1
    for i in range(len(hermite)):
        total *= filter(None, double_hermite[i])[0]
    return total

def integer_matrix(vectors):
    return [map(nint, v) for v in vectors]

def nint(x):
    return int(round(x))

#
# Test code
#

if __name__=='__main__':
    # Example from Wikipedia
    wpex = [[1, 0, -3, 0, 2, -8],
            [0, 1, 5, 0, -1, 4],
            [0, 0, 0, 1, 7, -9],
            [0, 0, 0, 0, 0, 0]]
    plimit = [1.0, 1.5849625007211563, 2.3219280948873622,
            2.8073549220576042, 3.4594316186372978, 3.7004397181410922]
    wpuv = unison_vectors(wpex,  plimit)
    for vector in [[-3, 5, -1, 0, 0, 0],
                   [3, 2, -1, 2, 1, 1],
                   [-4, 2, 0, 5, -2, -1]]:
        assert vector in wpuv, vector
    assert temper_out(wpuv) == regutils.lattice_reduction(wpex)[:3]

    # Check something obvious
    assert map(regutils.getRatio, TLLL([(-4,4,-1)], plimit)) == [(81,80)]

    # Known temperaments that can cause problems
    mystery = [[29, 46, 67, 81, 100, 107], [58, 92, 135, 163, 201, 215]]
    mystuv = unison_vectors(mystery, plimit)
    mystery_commas = map(regutils.getRatio, mystuv)
    for comma in [(196,195), (364,363), (441,440)]:
        # 1575:1573 also comes out
        # byt Pari gives 1001:1000 for this TLLL reduction
        assert comma in mystery_commas, comma
    assert temper_out(mystuv) == regutils.lattice_reduction(mystery)
    pajara = [[2, 3, 5, 6], [0, 1, -2, -2]]
    pajuv = unison_vectors(pajara, plimit) # spurious primes are fine
    assert map(regutils.getRatio, pajuv) == [(50,49), (64,63)]
    assert temper_out(pajuv) == regutils.lattice_reduction(pajara)
    comma = regutils.factorizeRatio(50, 49)[:4]
    assert temper_out([comma]) == [[2,0,0,1], [0,1,0,0], [0,0,1,1]]
    uvs = unison_vectors(temper_out([comma]), plimit)
    assert len(uvs) == 1
    assert regutils.getRatio(uvs[0]) == (50, 49)
    nums = 2, 3, 5, 11
    sublimit = [1.0, 1.5849625007211563, 2.3219280948873622, 3.4594316186372978]
    magick = [(19,30,44,66),(3,5,7,11)]
    magickuv = unison_vectors(magick, sublimit)
    magick_commas = [regutils.getRatio(uv, nums) for uv in magickuv]
    for comma in [(45, 44)]: #, (3125, 3072)]:
        assert  comma in magick_commas
    assert temper_out(magickuv) == regutils.lattice_reduction(magick)

