
#  Rank 2 temperament finding library
#  based on regular_minimal.py

###############################################################################
#
#  definitions
#
###############################################################################

cdef extern from "math.h":
    double log(double x)
    double sqrt(double x)
    long lrint(double x)

cdef extern from "stdlib.h":
    long abs(long n)
    void *malloc(int size)
    void free(void *ptr)

cdef double log2(double f):
  return log(f)/log(2)

cdef enum:
    prime_size = 11

cdef long primeNumbers[prime_size]

cdef int i, number
i=0
primes = []
for number in (2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31):
    primeNumbers[i] = number
    primes.append(log2(number))
    i = i + 1

# only mapping is set and used when building equal temperaments
cdef struct EqualTemperament:
    long *mapping
    double *errors
    double mean_error
    double error_variance

cdef struct MappingArray:
    int n_elements
    int vector_size
    long *elements

cdef struct ETArray:
    int n_elements
    int vector_size
    EqualTemperament *elements

cdef struct TemperamentStorage:
    int n_results
    int current_size
    double highest_badness

###############################################################################
#
#  Equal Temperaments
#
###############################################################################

def getEqualTemperaments(
        object primes, int nETs=20, int lowest=1, double cutoff=0.1):
    """Get nETs EqualTemperaments

    The primes are supplied as a list of logarithms of frequencies
    The cutoff is the largest allowable error multiplied by the step size
    """
    cdef double cprimes[prime_size]
    cdef int nNotes, size, nResults, i
    cdef long *mapping
    cdef MappingArray results

    size = len(primes)
    assert size <= prime_size
    set_double_array(primes, cprimes, size)

    init_mapping_array(&results, size, nETs)

    consist = []
    for nNotes from lowest <= nNotes < 100000:
        if not len(consist)<nETs:
            return consist[:nETs]
        nResults = get_rms_limited_mappings(&results, nNotes, cutoff/nNotes, cprimes)
        for i from 0 <= i < nResults:
            mapping = mapping_from_array(&results, i)
            if hcf(mapping, size)==1:
                consist.append(get_long_list(mapping, size))
        if nResults == nETs:
            break
        assert nResults < nETs, "overran results array!"

    free_mapping_array(&results)
    if nResults < nETs:
        raise TemperamentException, "not enough ETs found"
    return consist


cdef int get_rms_limited_mappings(
        MappingArray *ets, int nNotes, double maxRMS, double primes[]) except -1:
    """Find all ET mappings with the optimal weighted RMS less than that given

    Return the number of ETs found
    """
    assert nNotes>0
    assert maxRMS>0
    cdef double maxSum
    # convert the maximum RMS into the maximum sum of squares
    # this will steadily increase as the mapping increases
    maxSum = maxRMS*maxRMS*ets.vector_size
    return get_more_square_limited_mappings(
        ets, 0, 0, nNotes, maxSum, 0.0, 0.0, primes)

cdef int get_more_square_limited_mappings(
        MappingArray* ets, int nET, int oldSize,
        long nextGuess,
        double limit,
        double tot, double tot2,
        double primes[]):
    """recursive helper function for rmsLimitedMappings

    ets - the MappingArray being filled by results
    nET - the index in MappingArray of what we're looking at now
    oldSize - how much of seedMapping is so far used
    nextGuess - the next thing we want to add to the ET
    limit - how big the optimum sum square can get
    tot - running total of w
    tot2 - running total of w squared
    primes - logs of primes

    returns the new value for nET
    """
    cdef int i, j, results, moreResults, size
    cdef long nNotes, bestGuess, guess
    cdef double weightedSize, error
    cdef long *seedMapping, *newMapping

    size = ets.vector_size
    if nET >= ets.n_elements:
        return ets.n_elements ###
    seedMapping = mapping_from_array(ets, nET)
    i = oldSize + 1

    seedMapping[oldSize] = nextGuess
    nNotes = seedMapping[0]
    weightedSize = nextGuess/primes[oldSize]/nNotes
    tot = tot + weightedSize
    tot2 = tot2 + weightedSize*weightedSize
    error = i - tot*tot/tot2

    if error > limit:
        # not good enough
        return nET ###
    if i == size:
        # recursion stops here
        if (nET+1) < ets.n_elements:
            # duplicate this ET so that the next recursion can use it
            newMapping = mapping_from_array(ets, nET+1)
            for j from 0 <= j < size:
                newMapping[j] = seedMapping[j]
        return nET + 1 ###

    # now try the next prime along
    bestGuess = nint(primes[i]*nNotes/primes[0])
    results = get_more_square_limited_mappings(
        ets, nET, i, bestGuess, limit, tot, tot2, primes)
    guess = bestGuess
    while 1:
        guess = guess + 1
        moreResults = get_more_square_limited_mappings(
                ets, results, i, guess, limit, tot, tot2, primes)
        if moreResults == results:
            break
        results = moreResults
    guess = bestGuess
    while 1:
        guess = guess - 1
        moreResults = get_more_square_limited_mappings(
                ets, results, i, guess, limit, tot, tot2, primes)
        if moreResults == results:
            break
        results = moreResults

    return results

###############################################################################
#
#  Linear temperaments constructed from equal temperaments or mappings
#
###############################################################################

def getLinearTemperaments(ets,
        primes,
        highestComplexity,
        highestError,
        nResults=100):
    """Returns a list of temperaments as specified,
    ordered by badness
    """
    cdef double cprimes[prime_size]
    cdef long mapping[prime_size]
    cdef ETArray cETs
    cdef int i, size

    size = len(primes)
    assert size <= prime_size
    set_double_array(primes, cprimes, size)

    # convert the ETs into an array
    init_et_array(&cETs, size, len(ets))
    for i from 0 <= i < len(ets):
        assert len(ets[i])==size
        set_long_array(ets[i], mapping, size)
        init_equal_temperament(&cETs.elements[i], mapping, cprimes, size)

    result = get_linear_temperaments(
        &cETs, cprimes, highestComplexity, highestError, nResults)
    free_et_array(&cETs)
    return result

class LinearTemperament:
    """Rank 2 temperament"""
    def __init__(self, mapping1, mapping2, primes):
        cdef long generatorMapping[prime_size]
        cdef long cmapping1[prime_size], cmapping2[prime_size]
        cdef double limit[prime_size]
        cdef long generator, octaveDivision
        cdef int size

        size = len(primes)
        assert size <= prime_size
        set_double_array(primes, limit, size)

        set_long_array(mapping1, cmapping1, size)
        set_long_array(mapping2, cmapping2, size)

        set_mapping(cmapping1, cmapping2, size, generatorMapping)

        self.generatorMapping = get_long_list(generatorMapping, size)
        self.primes = primes
        self.primesByMelody = mapping1, mapping2

    def invariant(self):
        """something that can be used as a unique key"""
        cdef long generatorMapping[prime_size]
        cdef int size
        size = len(self.primes)
        assert size <= prime_size
        set_long_array(self.generatorMapping, generatorMapping, size)
        return get_invariant(generatorMapping, size)

    def complexity(self):
        """the range of complexities, a top-like measure"""
        cdef long mapping1[prime_size]
        cdef long mapping2[prime_size]
        cdef EqualTemperament cet1, cet2
        cdef double primes[prime_size]
        cdef int size
        cdef double result
        size = len(self.primes)
        assert size <= prime_size
        et1, et2 = self.primesByMelody
        set_long_array(et1, mapping1, size)
        set_long_array(et2, mapping2, size)
        set_double_array(self.primes, primes, size)
        init_equal_temperament(&cet1, mapping1, primes, size)
        init_equal_temperament(&cet2, mapping2, primes, size)

        result = get_complexity(&cet1, &cet2, primes, size)
        free_equal_temperament(&cet1)
        free_equal_temperament(&cet2)
        return result

    def optimalRMSError(self):
        """Return the optimal prime, RMS, weighted error"""
        cdef long mapping1[prime_size]
        cdef long mapping2[prime_size]
        cdef EqualTemperament cet1, cet2
        cdef double primes[prime_size]
        cdef int size
        cdef double result
        size = len(self.primes)
        assert size <= prime_size
        et1, et2 = self.primesByMelody
        set_long_array(et1, mapping1, size)
        set_long_array(et2, mapping2, size)
        set_double_array(self.primes, primes, size)
        init_equal_temperament(&cet1, mapping1, primes, size)
        init_equal_temperament(&cet2, mapping2, primes, size)

        result = optimal_rms_error(&cet1, &cet2, size)
        free_equal_temperament(&cet1)
        free_equal_temperament(&cet2)
        return result

    def contorted(self):
        cdef long mapping1[prime_size]
        cdef long mapping2[prime_size]
        cdef int size
        size = len(self.primes)
        assert size <= prime_size
        et1, et2 = self.primesByMelody
        set_long_array(et1, mapping1, size)
        set_long_array(et2, mapping2, size)
        return is_contorted(mapping1, mapping2, size)

    def __repr__(self, consonances=None):
        # send back useful stuff
        error = self.optimalRMSError() # evil side effect sets the basis
        et1, et2 = self.primesByMelody
        n, m = et1[0], et2[0]
        params = {
            'mapping' : self.generatorMapping,
            'melody' : self.primesByMelody,
            'complexity' : self.complexity(),
            'rmsError' : error*1200,
        }
        return """generator mapping:
%(mapping)s

mapping by steps:
%(melody)s

complexity measure: %(complexity).3f
RMS weighted error: %(rmsError).3f cents/octave""" % params


cdef object get_linear_temperaments(
        ETArray *ets,
        double primes[],
        double highestComplexity,
        double highestError,
        int nResults):
    """C implementation of getLinearTemperaments
    still used Python objects
    but takes equal temperaments and primes as arrays
    """
    cdef double width, badness, m12, etprod
    cdef long octaveDivision, generator
    cdef EqualTemperament *et1, *et2
    cdef int i, j, size
    cdef TemperamentStorage temperaments
    cdef object data, queue # Python object can't be in a C struct

    size = ets.vector_size

    init_storage(&temperaments, nResults)
    data, queue = {}, []
    for i from 0 <= i < ets.n_elements - 1:
        et1 = &ets.elements[i]
        for j from i < j < ets.n_elements:
            et2 = &ets.elements[j]
            try:
                etprod = et_product(et1, et2, primes)
                m12 = product_mean(et1, et2, size)
                badness = optimal_badness(et1, et2, m12, etprod)
                if good_enough(&temperaments, badness):
                    width = new_complexity(et1, et2, m12, etprod)
                    if width <= highestComplexity:
                        if badness/width < highestError and not is_contorted(
                                    et1.mapping, et2.mapping, size):
                            add_temperament(&temperaments, data, queue,
                                et1.mapping, et2.mapping,
                                badness, size)
            except DegenerateException:
                continue

    return retrieve_temperaments(queue, get_double_list(primes, size))

# C implementations of LinearTemperament methods ...

cdef void set_mapping(
       long mapping1[], long mapping2[], int size,
       long generatorMapping[]) except *:

    cdef long m, n
    cdef int i

    m, n = mapping1[0], mapping2[0]
    for i from 0 <= i < size:
        generatorMapping[i] = n*mapping1[i] - m*mapping2[i]

# new complexity for the Python interface
cdef double get_complexity(
        EqualTemperament *et1, EqualTemperament *et2,
        double primes[], int size):
    cdef double minVal, maxVal, e
    cdef int i

    return new_complexity(et1, et2, product_mean(et1, et2, size),
            et_product(et1, et2, primes))

# error calculation for the Python interface
cdef double optimal_rms_error(
        EqualTemperament *et1, EqualTemperament *et2, int size) except *:

    cdef double shrink1, shrink2, zmd_var1, zmd_var2, zmd_cov
    cdef double mean1, mean2, m12, top, bottom

    mean1 = et1.mean_error
    mean2 = et2.mean_error
    shrink1 = mean1 + 1.0
    shrink2 = mean2 + 1.0

    m12 = product_mean(et1, et2, size)

    zmd_cov = (m12 - mean1*mean2)/shrink1/shrink2
    zmd_var1 = et1.error_variance/shrink1/shrink1
    zmd_var2 = et2.error_variance/shrink2/shrink2
    top = zmd_var1*zmd_var2 - zmd_cov*zmd_cov
    bottom = top + zmd_var1 + zmd_var2 - zmd_cov - zmd_cov
    return sqrt(top/bottom)

# optimal calculation used by C
cdef double new_complexity(
        EqualTemperament *et1, EqualTemperament *et2,
        double m12, double etprod) except *:
    cdef double mean1, mean2, meanprod, meansq1, meansq2

    mean1 = et1.mean_error
    mean2 = et2.mean_error
    meanprod = m12 + mean1 + mean2
    meansq1 = et1.error_variance + mean1*mean1 + 2*mean1
    meansq2 = et2.error_variance + mean2*mean2 + 2*mean2
    return etprod*sqrt(
            meansq1*meansq2 - meanprod*meanprod
            + meansq1 + meansq2 - 2*meanprod)

cdef double optimal_badness(
        EqualTemperament *et1, EqualTemperament *et2,
        double m12, double etprod) except *:
    cdef double cov12

    cov12 = m12 - et1.mean_error * et2.mean_error
    return etprod*sqrt(et1.error_variance*et2.error_variance - cov12*cov12)

cdef double product_mean(
        EqualTemperament *et1, EqualTemperament *et2, int size):
    cdef int i
    cdef double s12
    s12 = 0.0
    for i from 0 <= i < size:
        s12 = s12 + et1.errors[i]*et2.errors[i]
    return s12/size

cdef double et_product(
        EqualTemperament *et1, EqualTemperament *et2,
        double primes[]):
    return et1.mapping[0]*et2.mapping[0]/primes[0]/primes[0]

cdef object get_invariant(long mapping[], int size):

    cdef object result
    cdef int i, sign

    result = []
    sign = 0
    for i from 1 <= i < size:
        if sign==0:
            if mapping[i] < 0:
                sign = -1
            elif mapping[i] > 0:
                sign = 1
        result.append(mapping[i]*sign)

    if sign==0:
        raise DegenerateException, "linearly dependent"

    return tuple(result) ###

cdef int is_contorted(
        long mapping1[], long mapping2[], int size) except -1:
    """Wedgie and hcf code mixed in together"""
    cdef long m, n, swap
    cdef int i, j

    m = 0
    for i from 0 <= i < size-1:
        for j from i < j < size:
            n = mapping1[i]*mapping2[j] - mapping1[j]*mapping2[i]
            while n:
                swap = n
                n = m%n
                m = swap
            if m==1 or m==-1:
                return 0 ###

    return 1

# temperament storage implementation ...

cdef void init_storage(TemperamentStorage *storage, int n_results) except * :
    assert storage != NULL
    storage.n_results = n_results
    storage.current_size = 0
    storage.highest_badness = 0.0

cdef int good_enough(TemperamentStorage *storage, double badness):
    return (storage.current_size < storage.n_results) or (
            badness < storage.highest_badness)

cdef void add_temperament(
        TemperamentStorage *storage, object data, object queue,
        long et1[], long et2[],
        double badness, int size) except *:
    cdef object key, newItem, mapping1, mapping2
    cdef long generatorMapping[prime_size]

    if (storage.current_size >= storage.n_results) and (
            badness > storage.highest_badness):
        return ###
    set_mapping(et1, et2, size, generatorMapping)
    key = get_invariant(generatorMapping, size)
    mapping1 = get_long_list(<long *> et1, len(primes))
    mapping2 = get_long_list(<long *> et2, len(primes))
    newItem = badness, mapping1, mapping2
    if key in data:
        oldItem = data[key]
        if newItem >= oldItem:
            return ###
        try:
            queue.remove(oldItem)
        except ValueError:
            pass
    data[key] = newItem
    queue.append(newItem)
    queue.sort()
    del queue[storage.n_results:]
    storage.highest_badness = queue[-1][0]
    storage.current_size = len(queue)

cdef object retrieve_temperaments(object queue, object primes):
    cdef object results, et1, et2, badness
    results = []
    for badness, et1, et2 in queue:
        results.append(LinearTemperament(et1, et2, primes))
    return results

###############################################################################
#
#  Utilities
#
###############################################################################

class TemperamentException(Exception):
  pass

class DegenerateException(TemperamentException):
  pass

cdef long hcf(long *numbers, int length):
    cdef long m, n, swap
    cdef int i
    m = 0
    for i from 0 <= i < length:
        n = numbers[i]
        if m==1 or m==-1:
            return 1 ###
        while n:
            swap = n
            n = m%n
            m = swap
    return abs(m)

# hacks to find out the size of things without access to sizeof
cdef long *long_ptr
cdef int long_size
long_size = <int>(long_ptr+1)-<int>long_ptr
cdef double *double_ptr
cdef int double_size
double_size = <int>(double_ptr+1)-<int>double_ptr
cdef EqualTemperament *et_ptr
cdef int et_size
et_size = <int>(et_ptr+1) - <int>et_ptr

cdef void init_equal_temperament(
        EqualTemperament* et, long *mapping, double *primes, int vector_size
        ) except *:
    cdef int i
    cdef double step_size, error, sum_errors, sum_square_errors

    et.mapping = <long *> malloc(vector_size * long_size)
    if et.mapping == NULL:
        raise MemoryError, "failed to allocate mapping"
    et.errors = <double *> malloc(vector_size * double_size)
    if et.errors == NULL:
        raise MemoryError, "failed to allocate errors"

    step_size = primes[0]/mapping[0]
    sum_errors = sum_squared_errors = 0.0
    for i from 0 <= i < vector_size:
        et.mapping[i] = mapping[i]
        error = step_size*mapping[i]/primes[i] - 1.0
        et.errors[i] = error
        sum_errors = sum_errors + error
        sum_square_errors = sum_square_errors + error*error

    et.mean_error = sum_errors/vector_size
    et.error_variance = sum_square_errors/vector_size - et.mean_error**2

cdef void init_et_array(
        ETArray* self, int vector_size, int n_elements) except *:
    cdef int i
    assert(self != NULL)
    self.vector_size = vector_size
    self.n_elements = n_elements
    self.elements = <EqualTemperament *> malloc(n_elements * et_size)
    if self.elements == NULL:
        raise MemoryError, "failed to allocate ET array"
    for i from 0 <= i < n_elements:
        self.elements[i].mapping = NULL
        self.elements[i].errors = NULL

cdef void init_mapping_array(
        MappingArray* self, int vector_size, int n_elements) except *:
    """return a two dimensional array of longs"""
    assert(self != NULL)
    self.vector_size = vector_size
    self.n_elements = n_elements
    self.elements = <long *> malloc(n_elements * vector_size * long_size)
    if self.elements == NULL:
        raise MemoryError, "failed to allocate ET array"

cdef long* mapping_from_array(MappingArray* self, int element) except NULL:
    assert(self != NULL)
    if element < 0 or element >= self.n_elements:
        raise IndexError, "array index out of range"
    return self.elements + element*self.vector_size

cdef void free_equal_temperament(EqualTemperament *et):
    if et.mapping != NULL:
        free(et.mapping)
    if et.errors != NULL:
        free(et.errors)

# note: this is never referred to so there's a memory leak
cdef void free_et_array(ETArray* self):
    cdef int i
    assert self != NULL
    assert self.elements != NULL
    for i from 0 <= i < self.n_elements:
        free_equal_temperament(&self.elements[i])
    free(self.elements)

# this neither
cdef void free_mapping_array(MappingArray* self):
    assert self != NULL
    assert self.elements != NULL
    free(self.elements)

cdef object get_long_list(long input[], int size):
    """Return a Python list from an array of longs"""
    cdef int i
    cdef object result
    result = []
    for i from 0 <= i < size:
        result.append(input[i])
    return result

cdef object get_double_list(double input[], int size):
    """Return a Python list from an array of doubles"""
    cdef int i
    cdef object result
    result = []
    for i from 0 <= i < size:
        result.append(input[i])
    return result

cdef void set_long_array(object input, long output[], int size) except *:
    """Set the C array to be like the Python list"""
    for i from 0 <= i < size:
        output[i] = input[i]

cdef void set_double_array(object input, double output[], int size) except *:
    """Set the C array to be like the Python list"""
    for i from 0 <= i < size:
        output[i] = input[i]

cdef long nint(double x):
    return lrint(x)
