; Regular temperament finding library.
; Based on regular_minimal.py

; -----------------------------------------------------------------------------
;
; Basics for defining the prime limits
;
; -----------------------------------------------------------------------------

(define (log2 f) (/ (log f) (log 2)))

; Intervals are measured in terms of 2:1 octaves.
; Sometimes, the first entry in a list of primes is taken to be the
; equivalence interval.
(define prime-numbers '(2 3 5 7 11 13 17 19 23 29 31))
(define primes (map log2 prime-numbers))

(define (limit n)
    (let more-primes ((basis primes) (nums prime-numbers))
        (cond
            ((null? nums) (display "Number is not a registered prime"))
            ((= (car nums) n) (list (car basis)))
            (else (cons (car basis) (more-primes (cdr basis) (cdr nums)))))))

; -----------------------------------------------------------------------------
;
; Utility funcitons
;
; -----------------------------------------------------------------------------

; nearest integer
(define (nint n) (inexact->exact (round n)))

; Highest common factor of a list of numbers.
; gcd should do this
; but it's only binary in TinyScheme
(define (hcf numbers)
    (let sub-hcf ((m 0) (others numbers))
        (if (or (null? others) (= m 1)) m
            (sub-hcf (gcd (car others) m) (cdr others)))))

; Arithmetic mean of a list.
(define (mean X) (/ (apply + X) (length X)))

; Statistical variance of a list
(define (variance X)
    (- (mean (map square X)) (square (mean X))))

; Statistical covariance of two lists
(define (covariance X Y)
    (- (mean (map * X Y)) (* (mean X) (mean Y))))

(define (square x) (* x x))

; Return a head list no larger than target-size
(define (trim-list target-size items)
    (cond ((< (length items) target-size) items)
          ((or (null? items) (zero? target-size)) '())
          (else (cons (car items)
                      (trim-list (- target-size 1) (cdr items))))))

; -----------------------------------------------------------------------------
;
; Equal Temperaments
;
; -----------------------------------------------------------------------------

; Return all ET mappings with the given number of notes and the
; optimal weighted RMS less than that given.
(define (limited-mappings n-notes max-std basis)
    (let (
          (limit (* max-std max-std (length basis)))
          (step-size (/ (car basis) n-notes)))
      (let more-mappings (
            (seed-mapping '()) ; the ET ampping so far
            (next-guess n-notes) ; the next thing to add to the ET
            (tot 0.0) (tot2 0.0) ; running total of w and w squared
            (more-primes basis)) ; primes for subsequent mappings
        (let* ((mapping (append seed-mapping (list next-guess)))
               (weighted-size (/ (* next-guess step-size) (car more-primes)))
               ; the next few lines determine the type of error
               (tot (+ tot weighted-size))
               (tot2 (+ tot2 (* weighted-size weighted-size)))
               (error (- (length mapping) (/ (* tot tot) tot2))))
            (cond
                ((> error limit) '()) ; too bad
                ((= (length mapping) (length basis))
                    ; end of the road
                    (if (= (hcf mapping) 1) (list mapping) '()))
                (else
                     (let ((best-guess (nint
                            (/ (cadr more-primes) step-size))))
                        (define (try-all guess op)
                            ; Try with this guess,
                            ; and keep op-ing one until you get one
                            ; that doesn't work.
                            ; The operator op is either + or -
                            (let ((results (
                                  more-mappings
                                      mapping (op guess 1) tot tot2
                                      (cdr more-primes))))
                              (if (> (length results) 0)
                                  (append results (try-all (op guess 1) op))
                                  results)))
                        ; Try the best guess,
                        ; and those above and below it.
                        (append (more-mappings
                                       mapping best-guess tot tot2
                                       (cdr more-primes))
                                (try-all best-guess +)
                                (try-all best-guess -)))))))))

; Gets n-ets equal temperament mappings.
; The basis is a list of prime intervals in terms of octaves.
; The cutoff is the largest allowable error multiplied by the step size.
; The first ET it looks at is "lowest" notes per octave.
(define (get-equal-temperaments basis n-ets cutoff lowest)
    (if (> lowest 10000)
      ; don't get too far into an infinite loop
      (display "Not enough ETs found")
      (let* (
          (mappings (limited-mappings lowest (/ cutoff lowest) basis))
          (n-found (length mappings)))
        (if (>= n-found n-ets)
            (trim-list n-ets mappings)
            (append mappings (get-equal-temperaments
                basis (- n-ets n-found) cutoff (+ lowest 1)))))))

; Apparatus to treat ETs as objects.
; They store the things we need for the rank 2 calculations.
(define (et.new mapping basis)
    (let* ((nNotes (car mapping))
           (octave (/ (car mapping) (car basis)))
           (weighted-err
               (map (lambda (steps p) (- (/ steps octave p) 1)) mapping basis)))
        (vector mapping
            (variance weighted-err)
            (mean weighted-err)
            weighted-err)))
(define (et.mapping et) (vector-ref et 0))
(define (et.var et)     (vector-ref et 1))
(define (et.mean et)    (vector-ref et 2))
(define (et.err et)     (vector-ref et 3))

; -----------------------------------------------------------------------------
;
; Finite queue.
; Keeps the best items in the right order
;
; -----------------------------------------------------------------------------

; capacity is the maximum number of items
; badnessfn is a function that takes a queue item
; and returns a number to determine the sorting
(define (queue.new capacity badnessfn)
    (vector '() capacity badnessfn 0)) ; the initial badness is arbitrary

(define (queue.items queue)         (vector-ref queue 0))
(define (queue.capacity queue)      (vector-ref queue 1))
(define (queue.badnessfn queue)     (vector-ref queue 2))
(define (queue.worst-badness queue) (vector-ref queue 3))
(define (queue.set-items! queue items) (vector-set! queue 0 items))
(define (queue.set-worst-badness! queue badness) (vector-set! queue 3 badness))
(define (queue.extract queue)
    (trim-list (queue.capacity queue) (queue.items queue)))

(define (queue.good-enough? queue badness)
    (or (< badness (queue.worst-badness queue))
        (< (length (queue.items queue)) (queue.capacity queue))))

; add an item
; keep the items stored with the best at the head
; can get overfull
(define (queue.add! queue item)
    (let* ((badfn (queue.badnessfn queue))
           (badness (badfn item)))
        (queue.set-items! queue
            (let check-next (
                    (capacity (queue.capacity queue))
                    (items (queue.items queue)))
                (cond ((zero? capacity) '())
                      ((null? items) (begin
                           (queue.set-worst-badness! queue badness)
                           (list item)))
                      ((< badness (badfn (car items)))
                         (if (= capacity 1)
                             (begin
                                 (queue.set-worst-badness! queue badness)
                                 (list item))
                             (cons item items)))
                      ((= capacity 1) (begin
                          (queue.set-worst-badness! queue (badfn (car items)))
                          items))
                      (else
                          (cons (car items)
                                (check-next (- capacity 1) (cdr items)))))))))

; remove an item
; may leave the worst badness too high, but better that than too low!
(define (queue.remove! queue item)
    (queue.set-items! queue
        (let check-next (
                (capacity (queue.capacity queue))
                (items (queue.items queue)))
            (cond ((or (null? items) (zero? capacity)) '())
                  ((eqv? item (car items)) (cdr items))
                  (else
                      (cons (car items)
                            (check-next (- capacity 1) (cdr items))))))))

; -----------------------------------------------------------------------------
;
; Rank 2 Temperaments
;
; -----------------------------------------------------------------------------

; apparatus to treat rank 2 temperaments as objects
(define (lt.new et1 et2 basis)
     ; Constructs a rank 2 temperament from two equal temperaments.
     ; basis -- the prime intervals in terms of octaves
     (let ((cov12 (- (mean (map * (et.err et1) (et.err et2)))
                     (* (et.mean et1) (et.mean et2))))
           (etprod (/ (* (car (et.mapping et1)) (car (et.mapping et2)))
                      (* (car basis) (car basis)))))
          (vector et1 et2 basis cov12 etprod)))
(define (lt.et1 lt)    (vector-ref lt 0))
(define (lt.et2 lt)    (vector-ref lt 1))
(define (lt.basis lt)  (vector-ref lt 2))
(define (lt.cov12 lt)  (vector-ref lt 3))
(define (lt.etprod lt) (vector-ref lt 4))

; scalar complexity
(define (lt.complexity lt)
    (let* ((var1 (et.var (lt.et1 lt)))
           (var2 (et.var (lt.et2 lt)))
           (mean1 (et.mean (lt.et1 lt)))
           (mean2 (et.mean (lt.et2 lt)))
           (meanprod (+ (lt.cov12 lt) (* mean1 mean2) mean1 mean2))
           (meansq1 (+ var1 (* mean1 mean1) (* 2 mean1)))
           (meansq2 (+ var2 (* mean2 mean2) (* 2 mean2))))
        (* (lt.etprod lt) (sqrt
                (- (+ (* meansq1 meansq2) meansq1 meansq2)
                   (+ (square meanprod) meanprod meanprod))))))

; The optimal weighted RMS error*complexity
(define (lt.optimal-badness lt)
    (let* ((var1 (et.var (lt.et1 lt)))
           (var2 (et.var (lt.et2 lt)))
           (sqbadness (- (* var1 var2) (square (lt.cov12 lt)))))
        (if (< sqbadness 0) 0 (* (lt.etprod lt) (sqrt sqbadness)))))

; first generator in cents with optimal tuning
(define (lt.g1 lt)
    (let* ((var1 (et.var (lt.et1 lt)))
           (var2 (et.var (lt.et2 lt)))
           (mean1 (et.mean (lt.et1 lt)))
           (mean2 (et.mean (lt.et2 lt)))
           (mw1 (+ 1 mean1))
           (mw2 (+ 1 mean2))
           (mw12 (+ 1 (lt.cov12 lt) (* mean1 mean2) mean1 mean2))
           (mw1sq (+ 1 var1 (* mean1 mean1) (* 2 mean1)))
           (mw2sq (+ 1 var2 (* mean2 mean2) (* 2 mean2))))
        (* 1200
            (/ (- (* mw1 mw2sq) (* mw2 mw12))
               (- (* mw1sq mw2sq) (* mw12 mw12))
               (car (et.mapping (lt.et1 lt)))))))

; second generator in cents with optimal tuning
(define (lt.g2 lt)
    (let* ((var1 (et.var (lt.et1 lt)))
           (var2 (et.var (lt.et2 lt)))
           (mean1 (et.mean (lt.et1 lt)))
           (mean2 (et.mean (lt.et2 lt)))
           (mw1 (+ 1 mean1))
           (mw2 (+ 1 mean2))
           (mw12 (+ 1 (lt.cov12 lt) (* mean1 mean2) mean1 mean2))
           (mw1sq (+ 1 var1 (* mean1 mean1) (* 2 mean1)))
           (mw2sq (+ 1 var2 (* mean2 mean2) (* 2 mean2))))
        (* 1200
            (/ (- (* mw2 mw1sq) (* mw1 mw12))
               (- (* mw1sq mw2sq) (* mw12 mw12))
               (car (et.mapping (lt.et2 lt)))))))

; The octave-equivalent generator mapping multiplied by
; the number of periods to the octave.
; Identical to the octave-equivalent wedgie.
; Normalized so that the first entry is always positive.
(define (lt.invariant lt)
    (let* ((mapping1 (et.mapping (lt.et1 lt)))
           (mapping2 (et.mapping (lt.et2 lt)))
           (m (car mapping1))
           (n (car mapping2))
           (genmap (map
                     (lambda (m1 m2) (- (* m m2) (* n m1)))
                     (cdr mapping1) (cdr mapping2))))
        (if (let first-positive? ((items genmap))
                (let ((item (car items)))
                    (cond ((< item 0) #f)
                          ((> item 0) #t)
                          (else (first-positive? (cdr items))))))
            genmap
            (map - genmap))))

; A regular temperament is contorted iff some tempered intervals don't
; represent any intervals from the set beign approximated.
; Also, iff the wedgie has a GCD other than 1 or -1.
; Here, I calculate the gcd of the wedgie all in one go.
; It's ugly, but it's more efficient than calculating the wedgie first
; because you can give up when the gcd hits 1.
(define (lt.contorted? lt)
    (not (= 1
        (let gcd-wedgie ((init-gcd 0)
                         (a (et.mapping (lt.et1 lt)))
                         (b (et.mapping (lt.et2 lt))))
            (if (or (= init-gcd 1) (null? a)) init-gcd
                (let ((ai (car a)) (bi (car b)))
                    (gcd-wedgie
                        (gcd init-gcd
                            (let loop ((m init-gcd) (a (cdr a)) (b (cdr b)))
                                (if (or (= m 1) (null? a)) m
                                    (let ((n (- (* ai (car b)) (* bi (car a)))))
                                        (loop (gcd m n) (cdr a) (cdr b))))))
                        (cdr a) (cdr b))))))))

; pseudo-objects for items to store in the queue
; (items for the queue, not queues)
(define (lt-queue.new lt badness)
    (cons (lt.invariant lt) (vector badness lt)))
(define (lt-queue.invariant lt-queue) (car lt-queue))
(define (lt-queue.badness   lt-queue) (vector-ref (cdr lt-queue) 0))
(define (lt-queue.lt        lt-queue) (vector-ref (cdr lt-queue) 1))

; Return a list of temperaments as specified, orderd by minimax badness.
; mappings -- equal temperament mappings used as the seed
; basis -- prime intervals in terms of octaves
; highest-complexity -- weighted as steps per octave, or whatever
; highest-error -- weighted and dimensionless
; n-results -- the maximum number of rank 2 temperaments to return
(define (get-linear-temperaments
          mappings basis
          highest-complexity highest-error
          n-results)
    (let ((temperaments (queue.new n-results lt-queue.badness))
          ; The results are stored in the "temperaments" container.
          ; Now convert the mappings into ET pseudo-objects.
          (ets (map (lambda (mapping) (et.new mapping basis)) mappings)))
        (define (try-lt! et1 et2)
            (let* ((lt (lt.new et1 et2 basis))
                   (width (lt.complexity lt)))
                (if (<= width highest-complexity)
                    (let* ((badness (lt.optimal-badness lt))
                           (error (/ badness width)))
                        (if (and (queue.good-enough? temperaments badness)
                                 (<= error highest-error)
                                 (not (lt.contorted? lt)))
                            (let* ((item (lt-queue.new lt badness))
                                   (dupe (assoc (lt-queue.invariant item)
                                                (queue.items temperaments))))
                                (if dupe
                                    ; only store a duplicated invariant if
                                    ; the temperament's better than the old one
                                    (if (< badness (lt-queue.badness dupe))
                                        (begin
                                            (queue.remove! temperaments dupe)
                                            (queue.add! temperaments item)))
                                    (queue.add! temperaments item))))))))
        ; Look at all pairs of equal temperaments.
        (let outer-loop ((ets ets))
            (if (null? ets)
                (map lt-queue.lt (queue.extract temperaments))
                (begin
                    (let ((et1 (car ets)))
                        (begin
                            (let inner-loop ((ets (cdr ets)))
                                (if (not (null? ets))
                                    (begin
                                        (try-lt! et1 (car ets))
                                        (inner-loop (cdr ets)))))))
                    (outer-loop (cdr ets)))))))

; -----------------------------------------------------------------------------
;
; Test code
;
; -----------------------------------------------------------------------------

; Duplicates the "new" output of regular.py.
; See comments there for details.

(define (round-float number decimal-places)
    (let ((multiplier
            (let pow-10 ((n decimal-places))
                (if (<= n 0) 1 (* 10 (pow-10 (- n 1)))))))
         (/ (round (* multiplier number)) multiplier)))

; make a list look like a Python tuple
(define (legacy-format a)
    (string-append "("
        (let loop ((items a))
            (if (null? items)
                ")"
                (string-append
                      (number->string (car items))
                      (if (null? (cdr items)) "" ", ")
                      (loop (cdr items)))))))

; print a list space separated without parens
(define (display-list items)
    (if (not (null? items))
        (begin
            (display (car items))
            (if (not (null? (cdr items))) (display " "))
            (display-list (cdr items)))))

; like display-list but cuts out the middle bit
(define (display-list-ends items)
    (if (> (length items) 30)
        (begin
            (display-list (trim-list 10 items))
            (display " ... ")
            (display-list (list-tail items (- (length items) 10))))
        (display-list items)))

(define (write-lts basis et-cutoff n-ets highest-complexity worst-error)
    (let* ((highest-complexity
             (* highest-complexity 0.125 (length basis) (length basis)))
           (et-cutoff (/ et-cutoff (sqrt (length basis))))
           (ets (get-equal-temperaments basis n-ets et-cutoff 1)))
        (begin
            (display-list-ends (map car ets)) (newline)
            (display-list (map legacy-format (map lt.invariant
                (get-linear-temperaments
                    ets basis highest-complexity worst-error 100))))
            (display " ") (newline) (newline))))

(define (tubulong-partial s)
    (/ (* s (- (square s) 1)) (sqrt (* (+ (square s) 1) 7.2))))

(define tubulong-frequencies (map tubulong-partial (list 3 4 5 6 7 8)))
(define tubulong-basis (map log2 tubulong-frequencies))

(define (test)
    (let ((limits (list 5 7 11 13 17 19 23 29 31)))
       (begin
          (display "RMS Optima") (newline) (newline)
          (for-each (lambda (prime)
              (begin
                  (if (< prime 10) (display " "))
                  (display prime)
                  (display "-limit") (newline)
                  (write-lts (limit prime) 0.17 20 2 0.002)
              )) limits)

          (display "Less good ETs") (newline) (newline)
          (for-each (lambda (prime)
              (begin
                  (if (< prime 10) (display " "))
                  (display prime)
                  (display "-limit") (newline)
                  (write-lts (limit prime) 0.22 20 2 0.005)
              )) limits)

          (display "Microtemperaments") (newline) (newline)
          (for-each (lambda (prime)
              (begin
                  (if (< prime 10) (display " "))
                  (display prime)
                  (display "-limit") (newline)
                  (write-lts (limit prime) 0.17 30 5 0.0005)
              )) limits)

          (for-each (lambda (f) (begin
                (display (round-float f 5)) (display " ")))
                tubulong-frequencies)
          (newline)

          (display "Octave-Equivalent Tubulongs") (newline) (newline)
          (let next-limit ((size 2))
              (if (<= size (length tubulong-basis))
              (begin
                  (write-lts
                      (cons 1.0 (trim-list size tubulong-basis))
                      0.17 20 2 0.002)
                  (next-limit (+ size 1)))))

          (display "Straight Tubulongs") (newline) (newline)
          (let next-limit ((size 3))
              (if (<= size (length tubulong-basis))
              (begin
                  (write-lts
                      (trim-list size tubulong-basis)
                      0.17 20
                      (/ 2 (car tubulong-basis))
                      0.002)
                  (next-limit (+ size 1))))))))

(define (big-test n-ets)
    (let ((limits (list 5 7 11 13 17 19 23 29 31)))
       (begin
          (display "RMS Optima") (newline) (newline)
          (for-each (lambda (prime)
              (begin
                  (if (< prime 10) (display " "))
                  (display prime)
                  (display "-limit") (newline)
                  (write-lts (limit prime) 0.2 n-ets 4 0.002)
              )) limits)

          (display "Inconsistent ETs") (newline) (newline)
          (for-each (lambda (prime)
              (begin
                  (if (< prime 10) (display " "))
                  (display prime)
                  (display "-limit") (newline)
                  (write-lts (limit prime) 0.3 n-ets 4 0.002)
              )) limits)

          (display "Microtemperaments") (newline) (newline)
          (for-each (lambda (prime)
              (begin
                  (if (< prime 10) (display " "))
                  (display prime)
                  (display "-limit") (newline)
                  (write-lts (limit prime) 0.2 n-ets 7 0.0005)
              )) limits))))

