#  Regular temperament finding library
#  based on temper.py
#  only does weighted prime errors
#  and generally more minimal
#
#  Hacked to work with the venerable Python 1.5.2

###############################################################################
#
#  definitions
#
###############################################################################

from math import log, sqrt, floor
import string

def log2(f):
  return log(f)/log(2)

primeNumbers = 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59
primeIntervals = map(log2, primeNumbers)


###############################################################################
#
#  Equal Temperaments
#
###############################################################################

def getEqualTemperaments(primes, nETs=20, lowest=1, cutoff=0.1):
    """Get nETs EqualTemperaments

    The primes are supplied as a list of logarithms of frequencies
    The cutoff is the largest allowable error multiplied by the step size
    """
    consist = []
    for nNotes in xrange(lowest, 10000):
        if not len(consist)<nETs:
            return consist[:nETs]
        for mapping in rmsLimitedMappings(nNotes, cutoff/nNotes, primes):
            if hcf(mapping)==1:
                consist.append(mapping)
    raise TemperamentException, "not enough ETs found"


def rmsLimitedMappings(nNotes, maxRMS, primes):
    """Return all ET mappings with the optimal weighted RMS less than that given"""
    assert nNotes>0, "asking for a negative number of notes"
    assert maxRMS>0, "asking for negative RMSs"

    # convert the maximum RMS into the maximum sum of squares
    # this will steadily increase as the mapping increases
    maxSum = maxRMS*maxRMS*len(primes)

    return moreSquareLimitedMappings([], nNotes, maxSum, 0.0, 0.0, primes)

def moreSquareLimitedMappings(seedMapping, nextGuess, limit, tot, tot2, primes):
    """recursive helper function for rmsLimitedMappings

    seedMapping - the ET mapping so far
    nextGuess - the next thing we want to add to the ET
    limit - how big the optimum sum square can get
    tot - running total of w
    tot2 - running total of w squared
    primes - logs of primes
    """
    mapping = seedMapping + [nextGuess]
    i = len(mapping)
    nNotes = mapping[0]
    weightedSize = nextGuess/primes[i-1]/nNotes
    tot = tot + weightedSize
    tot2 = tot2 + weightedSize**2
    error = i - tot**2/tot2
    if error>limit:
        # not good enough
        return [] ###
    if i==len(primes):
        # recursion stops here
        return [mapping] ###

    # now try the next prime along
    bestGuess = nint(primes[i]*nNotes/primes[0])
    results = moreSquareLimitedMappings(mapping, bestGuess, limit, tot, tot2, primes)
    guess = bestGuess
    while 1:
        guess = guess + 1
        moreResults = moreSquareLimitedMappings(
                mapping, guess, limit, tot, tot2, primes)
        if moreResults == []:
            break
        results = results + moreResults
    guess = bestGuess
    while 1:
        guess = guess - 1
        moreResults = moreSquareLimitedMappings(
                mapping, guess, limit, tot, tot2, primes)
        if moreResults == []:
            break
        results = results + moreResults

    return results

def etStats(mapping, primes):
    """useful things for calculating errors"""
    err = weightedErrors(mapping, primes)
    return mapping, err, mean(err), var(err)

def weightedErrors(mapping, primes):
    step = primes[0]/mapping[0]
    result = []
    for m, p in zip(mapping, primes):
        result.append(m/p*step - 1)
    return result

###############################################################################
#
#  Linear temperaments constructed from equal temperaments or mappings
#
###############################################################################


def getLinearTemperaments(ets,
        primes,
        highestComplexity,
        highestError,
        nResults=100):
    """Returns a list of temperaments as specified,
    ordered by badness
    """
    temperaments = TemperamentList(nResults)
    stats = []
    for et in ets:
        stats.append(etStats(et, primes))

    for i in range(len(ets)-1):
        stats1 = stats[i]
        for j in range(i+1, len(ets)):
            stats2 = stats[j]
            try:
                tuning = Rank2Temperament(stats1, stats2, primes)
                badness = tuning.optimalBadness()
                if temperaments.goodEnough(badness):
                    width = tuning.complexity()
                    if (width and width <= highestComplexity
                            and badness/width <= highestError
                            and not tuning.contorted()):
                        temperaments.add(tuning, badness)
            except DegenerateException:
                continue
            except ValueError:
                pass

    return temperaments.retrieve()

def LinearTemperament(mapping1, mapping2, primes):
    return Rank2Temperament(
            etStats(mapping1, primes),
            etStats(mapping2, primes),
            primes)

class Rank2Temperament:
    def __init__(self, stats1, stats2, primes):
        mapping1, err1, mean1, var1 = stats1
        mapping2, err2, mean2, var2 = stats2
        size = len(primes)
        assert len(err1)==len(err2) == len(mapping1)==len(mapping2) == size
        self.mappings = mapping1, mapping2
        self.errs = err1, err2
        self.means = mean1, mean2
        self.vars = var1, var2
        self.etprod = mapping1[0]*mapping2[0]/primes[0]**2
        self.primes = primes
        sumprod = 0.0
        for i in range(len(err1)):
            sumprod = sumprod + err1[i]*err2[i]
        self.meanprod = sumprod/size

    def optimalBadness(self):
        """error*complexity: the easiest kind to calculate"""
        var1, var2 = self.vars
        mean1, mean2 = self.means
        cov12 = self.meanprod - mean1*mean2
        return self.etprod*sqrt(var1*var2 - cov12*cov12)

    def complexity(self):
        var1, var2 = self.vars
        mean1, mean2 = self.means
        meanprod = self.meanprod + mean1 + mean2
        meansq1 = var1 + mean1*mean1 + 2*mean1
        meansq2 = var2 + mean2*mean2 + 2*mean2
        return self.etprod*sqrt(
                meansq1*meansq2 - meanprod*meanprod
                + meansq1 + meansq2 - 2*meanprod)

    def optimalRMSError(self):
        """Uses the true optimal scale stretch"""
        return self.optimalBadness()/self.complexity()

    def invariant(self):
        """something that can be used as a unique key"""
        mapping1, mapping2 = self.mappings
        m, n = mapping1[0], mapping2[0]
        mapping = []
        for i in range(1, len(mapping1)):
            mapping.append(m*mapping2[i] - n*mapping1[i])
        try:
            if filter(None, mapping)[0] < 0:
                result = []
                for x in mapping:
                    result.append(-x)
            else:
                result = mapping
        except IndexError:
            raise DegenerateException, "linearly dependent"
        return tuple(result)

    def contorted(self):
        # wedgie and hcf code mixed together
        a, b = self.mappings
        size = len(a)
        m = 0
        for i in range(size-1):
            for j in range(i+1, size):
                n = a[i]*b[j] - a[j]*b[i]
                while n:
                    m, n = n, m%n
                if m==1 or m==-1:
                    return 0
        return 1

    # these last methods are not required for the search
    def canonicalMapping(self):
        """return the period and generator mapping"""
        primesByMelody = apply(zip, self.mappings)
        m, n = primesByMelody[0]
        mGen, nGen, octave = extendedEuclid(m, n)
        if mGen+nGen < 0:
            mGen = mGen + m/octave
            nGen = nGen + n/octave
        assert m*nGen-n*mGen == octave, "generators wrong"
        genConv1, genConv2 = -n/octave, m/octave
        periodMapping = []
        generatorMapping = []
        for axis1, axis2 in primesByMelody:
            generatorMapping.append(genConv1*axis1 + genConv2*axis2)
            periodMapping.append(nGen*axis1 - mGen*axis2)
        return periodMapping, generatorMapping

    def optimalCanonicalGenerators(self):
        sw0 = sw1 = sw02 = sw12 = sw01 = 0.0

        periodMapping, generatorMapping = self.canonicalMapping()
        for i in range(len(self.primes)):
            m0, m1, p = periodMapping[i], generatorMapping[i], self.primes[i]
            w0, w1 = m0/p, m1/p
            sw0 = sw0 + w0
            sw1 = sw1 + w1
            sw02 = sw02 + w0**2
            sw12 = sw12 + w1**2
            sw01 = sw01 + w0*w1

        d = sw02*sw12 - sw01**2 # the determinant of the wNN matrix
        if d==0.0:
            raise DegenerateException, "null generator mapping"

        return (sw0*sw12 - sw1*sw01)/d, (sw1*sw02 - sw0*sw01)/d

class TemperamentList:
    def __init__(self, capacity):
        self.capacity = capacity
        self.data = {} # list of the things we've seen before
        self.queue = [] # the best ETs sorted by badness

    def goodEnough(self, badness):
        return len(self.queue)<self.capacity or badness<self.queue[-1][0]

    def add(self, temperament, badness):
        key = temperament.invariant()
        newItem = badness, temperament
        if self.data.has_key(key):
            oldItem = self.data[key]
            if newItem >= oldItem:
                return ###
            try:
                self.queue.remove(oldItem)
            except ValueError:
                pass
        self.data[key] = newItem
        self.queue.append(newItem)
        self.queue.sort()
        del self.queue[self.capacity:]

    def retrieve(self):
        result = []
        for badness, temperament in self.queue:
            result.append(temperament)
        return result

def tuningMap(mapping, generators):
    """Returns the tuning map for the given generators.
    Matching units.
    """
    result = []
    g0, g1 = generators
    mapping0, mapping1 = mapping
    for m0, m1 in zip(mapping0, mapping1):
        result.append(m0*g0 + m1*g1)
    return result

###############################################################################
#
#  Utilities
#
###############################################################################

class TemperamentException(Exception):
  pass

class DegenerateException(TemperamentException):
  pass

def mean(x):
    total = 0.0
    for each in x:
        total = total + each
    return total/len(x)

def var(x):
    """statistical variance"""
    size = len(x)
    tot2 = 0.0
    for each in x:
        tot2 = tot2 + each*each
    return tot2/size - mean(x)**2

def hcf(numbers):
    m = 0
    for n in numbers:
        while n:
            m, n = n, m%n
        if m==1:
            return 1
    return abs(m)

def extendedEuclid(f, g):
    """Return two generator sizes and the octave division
    from the sizes of two equal temperaments
    """
    # Uses the extended Euclidian algorithm
    # From von zur Gathen & Gerhard, p.47

    assert f>0 and g>0, "Bad arguments to extendedEuclid()"
    lastR, nextR = f, g
    lastS, lastT, nextS, nextT = 1, 0, 0, -1

    while 1:
        q, r = divmod(lastR, nextR)
        if r == 0:
            return nextT, nextS, nextR ###
        assert r>0, "extendedEuclid() internally inconsistent"

        nextR, lastR = r, nextR
        nextS, lastS = lastS - q*nextS, nextS
        nextT, lastT = lastT - q*nextT, nextT

def nint(x):
    return int(floor(x+0.5))

def zip(list1, list2):
    """simplified version of a built-in function for later versions"""
    size = len(list1)
    assert size==len(list2), "inconsistent sizes to zip"
    result = []
    for i in range(size):
        result.append((list1[i], list2[i]))
    return result

###############################################################################
#
#  Test code
#
###############################################################################

# originally duplicated the output of write_temper.ml
# now based on write_temper_rms.py
# only does RMS stuff because TOP optimization is so slow

def write_lts(primes, etCutoff=0.17,
        nETs=20, highestComplexity=2, worstError=0.002):
    # make adjustments
    highestComplexity = highestComplexity * len(primes)**2 * 0.125
    etCutoff = etCutoff / sqrt(len(primes))

    ets = getEqualTemperaments(primes, nETs=nETs, cutoff=etCutoff)
    octaves = []
    if len(ets)>30:
        for temp in ets[:10]:
            print temp[0],
        print '...',
        for temp in ets[-10:]:
            print temp[0],
        print
    else:
        for et in ets:
            octaves.append(str(et[0]))
        print string.join(octaves, ' ')

    for lt in getLinearTemperaments(ets, primes, highestComplexity, worstError):
        print lt.invariant(),
    print "\n"


tubulongFrequencies = []
for s in range(3,9):
    tubulongFrequencies.append(float(s*(s*s - 1)) / sqrt((s*s + 1) * 7.2))

tubulongPrimes = map(log2, tubulongFrequencies)

def test():
    primes = primeIntervals
    limits = range(2, 11) # indexes primes (or primeNumbers)
    print "RMS Optima\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1])

    print "Less good ETs\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1], etCutoff=0.22, worstError=0.005)

    print "Microtemperaments\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1], nETs=30, highestComplexity=5, worstError=5e-4)

    for f in tubulongFrequencies:
      print "%.5f"%f,
    print ''

    print "Octave-Equivalent Tubulongs\n"
    for limit in range(2, len(tubulongPrimes)+1):
        write_lts(([1.0]+tubulongPrimes)[:limit+1],  worstError = 0.002)

    print "Straight Tubulongs\n"
    for limit in range(2, len(tubulongPrimes)):
        write_lts(tubulongPrimes[:limit+1], worstError = 0.002,
                highestComplexity=2/tubulongPrimes[0])

def bigTest(nETs=300):
    primes = primeIntervals
    limits = range(2, 11) # indexes primes (or primeNumbers)
    print "RMS Optima\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1], nETs=nETs, etCutoff=0.2,
                highestComplexity=4, worstError=0.002)

    print "Inconsistent ETs\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1], nETs=nETs, etCutoff=0.3,
                highestComplexity=4, worstError=0.002)

    print "Microtemperaments\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1], nETs=nETs, etCutoff=0.2,
                highestComplexity=7, worstError=5e-4)

if __name__=='__main__':
    import sys
    if len(sys.argv)>1 and sys.argv[1].startswith('big'):
        if len(sys.argv)>2:
            bigTest(nETs=int(sys.argv[2]))
        else:
            bigTest()
    else:
        test()
