#  Regular temperament finding library
#  based on regular.py
#  only does RMS Tenney-weighted prime errors
#  and generally more minimal

from __future__ import division

###############################################################################
#
#  definitions
#
###############################################################################

from math import log, sqrt

def log2(f):
  return log(f)/log(2)

# Intervals are measured in terms of octaves (2:1)
# In general, the first entry in a list of primes is taken to be
# the equivalence interval.
primeNumbers = (2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31)
primes = map(log2, primeNumbers)

###############################################################################
#
#  Equal Temperaments
#
###############################################################################

def getEqualTemperaments(primes, nETs=20, cutoff=0.1):
    """Gets nETs equal temperament mappings.
    The primes are supplied as a list of logarithms of frequencies
    The cutoff is the largest allowable error multiplied by the step size
    """
    ets = []
    # Use a for loop over an arbitrarily large range
    # instead of an infinite loop
    # so that if the cutoff is too strict the program doesn't hang
    for nNotes in xrange(1, 10000):
        for mapping in rmsLimitedMappings(nNotes, cutoff/nNotes, primes):
            if hcf(mapping)==1:
                ets += [mapping]
                if len(ets) >= nETs:
                    return ets ###
    raise TemperamentException, "not enough ETs found"


def rmsLimitedMappings(nNotes, maxRMS, primes):
    """Return all ET mappings with the given number of notes and the
    optimal weighted RMS less than that given.
    """
    # convert the maximum RMS into the maximum sum of squares
    # this will steadily increase as the mapping increases
    limit = maxRMS*maxRMS*len(primes)

    def moreSquareLimitedMappings(seedMapping, nextGuess, tot, tot2):
        """seedMapping - the ET mapping so far
        nextGuess - the next thing we want to add to the ET
        tot - running total of w
        tot2 - running total of w squared
        """
        mapping = seedMapping + [nextGuess]
        i = len(mapping)
        nNotes = mapping[0]
        weightedSize = nextGuess/primes[i-1]/nNotes
        tot += weightedSize
        tot2 += weightedSize**2
        error = i - tot**2/tot2
        if error>limit:
            # not good enough
            return [] ###
        if i==len(primes):
            # recursion stops here
            return [mapping] ###

        # now try the next prime along
        # First check the nearest mapping of the prime.
        bestGuess = nint(primes[i]*nNotes/primes[0])
        results = moreSquareLimitedMappings(mapping, bestGuess, tot, tot2)

        # Then try the mappings above and below the nearest one.
        # Carry on in each direction until the limit's broken.
        guess = bestGuess
        while 1:
            guess += 1
            moreResults = moreSquareLimitedMappings(mapping, guess, tot, tot2)
            if moreResults == []:
                break
            results += moreResults
        guess = bestGuess
        while 1:
            guess -= 1
            moreResults = moreSquareLimitedMappings(mapping, guess, tot, tot2)
            if moreResults == []:
                break
            results += moreResults

        return results

    return moreSquareLimitedMappings([], nNotes, 0.0, 0.0)

def weightedErrors(mapping, primes):
    """Return the error of each prime interval
    scaled so that the optimal weighted size is 1
    and tuned so that the first prime is just.
    """
    assert len(mapping)==len(primes)
    step = primes[0]/mapping[0]
    return [m/p*step - 1 for m, p in zip(mapping, primes)]

###############################################################################
#
#  Linear temperaments constructed from equal temperaments or mappings
#
###############################################################################

def getLinearTemperaments(ets,
        primes,
        highestComplexity,
        highestError,
        nResults=100):
    """Returns a list of temperaments as specified,
    ordered by minimax badness.

    ets -- A list of mappings of equal temperaments that
           get paired off to produce rank 2 temperaments.
    primes -- The prime intervals in terms of octaves.
    highestComplexity -- weighted as steps per octave, or whatever.
    highestError -- weighted and dimensionless.
    error -- The function of weighted primes used to calculate the error.
             Some of these are imported from et.py
    nResults -- The maximum number of rank 2 temperaments to return.
    """
    temperaments = TemperamentList() # container for the results
    for i in range(len(ets)-1):
        et1 = ets[i]
        for j in range(i+1, len(ets)):
            et2 = ets[j]
            try:
                tuning = LinearTemperament(et1, et2, primes)
                width = tuning.complexity()
                if width <= highestComplexity:
                    badness = tuning.optimalBadness()
                    if badness/width < highestError and not tuning.contorted():
                        temperaments.add(tuning, badness)
            except ValueError:
                # Means the square root of a negative number was attempted.
                # This is an error on x86 chips.
                # It means the temperament has a very low error, maybe zero.
                # Whatever, we can do without it.
                continue

    return temperaments.retrieve()[:nResults]

class LinearTemperament:
    """Rank 2 temperament"""
    def __init__(self, mapping1, mapping2, primes):
        """Constructs a rank 2 temperament from two equal temperaments.
        mapping1 -- The mapping of prime intervals to scale steps
                    for the first equal temperament.
        mapping2 -- The mapping for the other equal temperament.
        primes   -- The prime intervals in terms of octaves.
        """
        self.mappings = mapping1, mapping2
        err1 = weightedErrors(mapping1, primes)
        err2 = weightedErrors(mapping2, primes)

        # These things are used in the badness and complexity functions.
        self.var1, self.var2 = var(err1), var(err2)
        self.means = mean(err1), mean(err2)
        self.cov12 = cov(err1, err2)
        self.etprod = mapping1[0]*mapping2[0]/primes[0]/primes[0]

    def optimalBadness(self):
        """The optimal weighted RMS error*complexity"""
        return self.etprod*sqrt(self.var1*self.var2 - self.cov12**2)

    def complexity(self):
        """scalar complexity"""
        mean1, mean2 = self.means
        meanprod = self.cov12 + mean1*mean2 + mean1 + mean2
        meansq1 = self.var1 + mean1*mean1 + 2*mean1
        meansq2 = self.var2 + mean2*mean2 + 2*mean2
        return self.etprod*sqrt(
                meansq1*meansq2 - meanprod*meanprod
                + meansq1 + meansq2 - 2*meanprod)

    def invariant(self):
        """The octave-equivalent generator mapping
        multiplied by the number of periods to the octave.
        Identical to the octave-equivalent wedgie.
        Normalized so that the first non-zero entry is always positive.
        """
        # This is where we still need the canonical generator mapping
        mapping1, mapping2 = self.mappings
        m, n = mapping1[0], mapping2[0]
        oewed = [m*mapping2[j] - n*mapping1[j] for j in range(1, len(mapping1))]
        if filter(None, oewed)[0]<0:
            return tuple([-x for x in oewed])
        else:
            return tuple(oewed)

    def contorted(self):
        """Returns true iff some tempered intervals don't represent
        any intervals from the set being approximated.
        """
        return hcf(wedgie(*self.mappings)) != 1

class TemperamentList:
    """Used for the containers temperaments are stored in during searches.
    This is designed for simplicity rather than efficiency.
    """
    def __init__(self):
        self.data = {} # temperaments indexed by invariant
    def add(self, temperament, badness):
        key = temperament.invariant()
        if key in self.data:
            if badness > self.data[key][0]:
                # make sure to keep the best temperament for a given invariant
                return ###
        self.data[key] = badness, temperament
    def retrieve(self):
        """Returns the temperaments as a list sorted by badness"""
        decorated = self.data.values()
        decorated.sort()
        return [temperament for badness, temperament in decorated]

###############################################################################
#
#  Utilities
#
###############################################################################

class TemperamentException(Exception):
  pass

def mean(x):
    """The arithmetic mean of a sequence"""
    total = 0.0
    for each in x:
        total += each
    return total/len(x)

def var(x):
    """The statistical variance of a sequence"""
    size = len(x)
    tot2 = 0.0
    for each in x:
        tot2 += each*each
    return tot2/size - mean(x)**2

def cov(x, y):
    """The statistical covariance of two sequences"""
    size = len(x)
    assert size==len(y)
    totab = 0.0
    for a, b in zip(x, y):
        totab += a*b
    return totab/size - mean(x)*mean(y)

def hcf(numbers):
    """The highest common factor of a sequence"""
    m = 0
    for n in numbers:
        if m==1 or m==-1:
            return 1
        while n:
            m, n = n, m%n
    return abs(m)

def nint(x):
    """The nearest integer to a given float"""
    return int(round(x))

def wedgie(a, b):
    """Simple way of generating the wedgie of two sequences"""
    size = len(a)
    assert size==len(b)
    for i in range(size-1):
        for j in range(i+1, size):
            yield a[i]*b[j] - a[j]*b[i]

###############################################################################
#
#  Test code
#
###############################################################################

# Duplicates the "new" output of regular.py.
# See comments there for details.

def write_lts(primes, etCutoff=0.17,
        nETs=20, highestComplexity=2, worstError=0.002):
    # make adjustments
    highestComplexity *= len(primes)**2 * 0.125
    etCutoff /= sqrt(len(primes))

    ets = getEqualTemperaments(primes, nETs=nETs, cutoff=etCutoff)
    print ' '.join([str(et[0]) for et in ets])

    for lt in getLinearTemperaments(ets, primes, highestComplexity, worstError):
        print lt.invariant(),
    print "\n"

tubulongFrequencies = [
    s*(s*s - 1) / sqrt((s*s + 1) * 7.2)
    for s in range(3,9)]

tubulongPrimes = map(log2, tubulongFrequencies)

def test():
    limits = range(2, 11) # indexes primes (or primeNumbers)
    print "RMS Optima\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1])

    print "Less good ETs\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1], etCutoff=0.22, worstError=0.005)

    print "Microtemperaments\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_lts(primes[:limit+1], nETs=30, highestComplexity=5, worstError=5e-4)

    for f in tubulongFrequencies:
      print "%.5f"%f,
    print ''

    print "Octave-Equivalent Tubulongs\n"
    for limit in range(2, len(tubulongPrimes)+1):
        write_lts(([1.0]+tubulongPrimes)[:limit+1],  worstError = 0.002)

    print "Straight Tubulongs\n"
    for limit in range(2, len(tubulongPrimes)):
        write_lts(tubulongPrimes[:limit+1], worstError = 0.002,
                highestComplexity=2/tubulongPrimes[0])

if __name__=='__main__':
    test()

