# Regular temperament library
# Does approximate octave-equivalent errors
# using statistical functions
from __future__ import generators

import et
from math import sqrt
try:
    from itertools import izip
except ImportError:
    izip = zip

###############################################################################
#
#  Tweaks to equal temperaments
#
###############################################################################

def cacheErrors(et):
    """Save some error results in the equal temperament instance"""
    err = et.weightedErrors()
    et.var = var(err)
    et.meanerr = mean(err)
    et.err = err[1:]

###############################################################################
#
#  Linear temperaments constructed from equal temperaments or mappings
#
###############################################################################

def getLinearTemperaments(ets,
        highestComplexity,
        highestError,
        nResults=100,
        temperamentFactory=None):
    """Returns a list of temperaments as specified,
    ordered by badness
    """
    # this is used by other modules to borrow this function
    if temperamentFactory is None:
        temperamentFactory = LinearTemperament
    temperaments = TemperamentList(nResults)
    for et in ets:
        cacheErrors(et)
    tuning = LinearTemperament(ets[0], ets[1]) #re-use this object
    for i in range(len(ets)-1):
        et1 = ets[i]
        for j in range(i+1, len(ets)):
            et2 = ets[j]
            tuning.setup(et1, et2)
            try:
                badness = tuning.optimalBadness()
                if temperaments.goodEnough(badness):
                    complexity = tuning.complexity()
                    if (complexity and complexity <= highestComplexity
                        and badness/complexity <= highestError
                        and not tuning.contorted()):
                            temperaments.add(tuning, badness)
            except ValueError:
                pass

    return [temperamentFactory(et1, et2) for et1, et2 in temperaments.retrieve()]

def Temperament(primes, m, n=None):
    et1 = et.BestET(m, primes)
    if n is None:
        return et1
    et2 = et.BestET(n, primes)
    return LinearTemperament(et1, et2)

temperament = Temperament

class LinearTemperament(et.PrimeWeighted):
    """Rank 2 temperament"""
    def __init__(self, et1, et2):
        """Note different constructor to other libraries"""
        try:
            self.setup(et1, et2)
        except AttributeError:
            cacheErrors(et1)
            cacheErrors(et2)
            self.setup(et1, et2)

    def setup(self, et1, et2):
        """In a separate method so we can repeat for an AttributeError"""
        self.ets = et1, et2
        self.primes = et1.primes
        size = len(et1.primes)
        assert size==len(et2.primes)
        assert len(et1.err)==len(et2.err)

        # pre-calcualate these things because they're used all over
        self.var1, self.var2 = et1.var, et2.var
        # calculate the covariance here for efficiency
        sumab = 0.0
        for a, b in izip(et1.err, et2.err):
            sumab += a*b
        self.cov12 = sumab/size - et1.meanerr*et2.meanerr
        self.etprod = et1.mapping[0]*et2.mapping[0]/self.primes[0]**2

    def optimalError(self):
        top = self.var1*self.var2 - self.cov12**2
        bottom = self.var1+self.var2 - 2*self.cov12
        return sqrt(top/bottom)

    def complexity(self):
        return self.etprod*sqrt(self.var1 + self.var2 - 2*self.cov12)

    def optimalBadness(self):
        """error*complexity: the easiest kind to calculate"""
        return self.etprod*sqrt(self.var1*self.var2 - self.cov12**2)

    def wedgieBadness(self):
        et1, et2 = self.ets
        return self.etprod*std(
               list(wedgie(et1.err, et2.err)))/sqrt(len(self.primes))

    def win(self):
        """wedgie sine"""
        et1, et2 = self.ets
        return sqrt(norm(wedgie(et1.err, et2.err))/norm(et1.err)/norm(et2.err))

    def invariant(self):
        """This is where we still need the canonical generator mapping"""
        et1, et2 = self.ets
        # mapping made explicit for speed
        m, n = et1.mapping[0], et2.mapping[0]
        oewed = [m*et2.mapping[j] - n*et1.mapping[j]
                for j in range(1, len(et1.mapping))]
        if filter(None, oewed)[0]<0:
            return tuple([-x for x in oewed])
        else:
            return tuple(oewed)

    def optimize(self):
        self.tuning = (self.var2 - self.cov12)/(
                self.var1 + self.var2 - 2*self.cov12)

    def error(self):
        return std(self.weightedErrors())

    def tuningMap(self):
        assert hasattr(self, "tuning"), "Must have a tuning" 
        t1 = self.ets[0].tuningMap()
        t2 = self.ets[1].tuningMap()
        g1 = self.tuning
        g2 = (1-self.tuning)
        return [g1*m1+g2*m2 for m1, m2 in zip(t1, t2)]

    def characteristicOctave(self):
        """Strange way of turning the complexity calculation backwards.
        Gives a number that's connected with an equal temperament
        """
        return sqrt(self.complexity()/self.error())

    def contorted(self):
        return et.hcf(wedgie(self.ets[0].mapping, self.ets[1].mapping)) != 1

    def cor(self):
        """Statistical correlation of the equal temperaments"""
        return self.cov12/sqrt(self.var1*self.var2)

    def rin(self):
        """counterpart to cor"""
        return sqrt(1 - self.cov12**2/(self.var1*self.var2))

    def sin(self):
        """Sine of angle between equal temperaments.

        Weighted mappings in euclidian space.
        """
        # approximate complexities
        comp0 = self.ets[0].mapping[0]*(1+self.ets[0].meanerr)
        comp1 = self.ets[1].mapping[0]*(1+self.ets[1].meanerr)
        return self.complexity()/comp0/comp1


    def __repr__(self):
        # send back useful stuff
        params = {
            'invariant' : self.invariant(),
            'et1' : self.ets[0],
            'et2' : self.ets[1],
            'complexity': self.complexity(),
            'error' : self.optimalError()*1200,
        }
        return """generator mapping:
%(invariant)s

mapping:
%(et1)s
%(et2)s

complexity: %(complexity).3f
optimal error: %(error).3f cents/octave""" % params

class TemperamentList:
    def __init__(self, capacity):
        self.capacity = capacity
        self.data = {} # list of the things we've seen before
        self.queue = [] # the best ETs sorted by badness

    def goodEnough(self, badness):
        return len(self.queue)<self.capacity or badness<self.queue[-1][0]

    def add(self, temperament, badness):
        key = temperament.invariant()
        newItem = badness, temperament.ets
        if key in self.data:
            oldItem = self.data[key]
            if newItem >= oldItem:
                return ###
            try:
                self.queue.remove(oldItem)
            except ValueError:
                pass
        self.data[key] = newItem
        self.queue.append(newItem)
        self.queue.sort()
        del self.queue[self.capacity:]

    def retrieve(self):
        return [ets for badness, ets in self.queue]

###############################################################################
#
#  Useful functions
#
###############################################################################

def mean(x):
    total = 0.0
    for each in x:
        total += each
    return total/len(x)

def std(x):
    """standard deviation"""
    return sqrt(var(x))

def var(x):
    """statistical variance"""
    size = len(x)
    tot = tot2 = 0.0
    for each in x:
        tot += each
        tot2 += each*each
    return (tot2 - tot*tot/size)/size

def norm(x):
    tot = 0.0
    for each in x:
        tot += each*each
    return et.sqrt(tot)

def wedgie(a, b):
    """Simple way of generating the wedgie of two sequences"""
    size = len(a)
    assert size==len(b)
    for i in range(size-1):
        for j in range(i+1, size):
            yield a[i]*b[j] - a[j]*b[i]

###############################################################################
#
#  Test code
#
###############################################################################

# adapted from regular_wedgie.py

def write_lts(primes, etCutoff=0.17,
        nETs=20, highestComplexity=2, worstError=0.002,
        error=et.stdError, func=getLinearTemperaments):
    # make adjustments
    highestComplexity *= len(primes)**2 * 0.125
    etCutoff /= sqrt(len(primes))

    ets = et.getEqualTemperaments(
            primes, nETs=nETs, cutoff=etCutoff, error=error)
    if len(ets)>30:
        for temp in ets[:10]:
            print temp[0],
        print '...',
        for temp in ets[-10:]:
            print temp[0],
        print
    else:
        print ' '.join([str(temp[0]) for temp in ets])

    for lt in func(ets, highestComplexity, worstError):
        print lt.invariant(),
    print "\n"

tubulongFrequencies = [
    float(s*(s*s - 1)) / sqrt((s*s + 1) * 7.2)
    for s in range(3,9)]

tubulongPrimes = map(et.log2, tubulongFrequencies)

# func and error are so other modules can use this
def test(func=getLinearTemperaments, error=et.stdError):
    limits = range(2, 11) # indexes primes (or primeNumbers)
    print "RMS Optima\n"
    for limit in limits:
        print "%2i-limit" % et.primeNumbers[limit]
        write_lts(et.primes[:limit+1], func=func, error=error)

    print "Less good ETs\n"
    for limit in limits:
        print "%2i-limit" % et.primeNumbers[limit]
        write_lts(et.primes[:limit+1], etCutoff=0.22, worstError=0.005,
                func=func, error=error)

    print "Microtemperaments\n"
    for limit in limits:
        print "%2i-limit" % et.primeNumbers[limit]
        write_lts(et.primes[:limit+1], nETs=30,
                highestComplexity=5, worstError=5e-4, func=func, error=error)

    for f in tubulongFrequencies:
      print "%.5f"%f,
    print ''

    print "Octave-Equivalent Tubulongs\n"
    for limit in range(2, len(tubulongPrimes)+1):
        write_lts([1.0]+tubulongPrimes[:limit], worstError = 0.002,
                func=func, error=error)

    print "Straight Tubulongs\n"
    for limit in range(2, len(tubulongPrimes)):
        write_lts(tubulongPrimes[:limit+1], worstError = 0.002,
                highestComplexity=2/tubulongPrimes[0],
                func=func, error=error)

def bigTest(nETs=300, func=getLinearTemperaments, error=et.stdError):
    limits = range(2, 11)
    print "RMS Optima\n"
    for limit in limits:
        print "%2i-limit"%et.primeNumbers[limit]
        write_lts(et.primes[:limit+1], nETs=nETs, etCutoff=0.2,
                highestComplexity=4, worstError=0.002,
                func=func, error=error)

    print "Inconsistent ETs\n"
    for limit in limits:
        print "%2i-limit"%et.primeNumbers[limit]
        write_lts(et.primes[:limit+1], nETs=nETs, etCutoff=0.3,
                highestComplexity=4, worstError=0.002, func=func, error=error)

    print "Microtemperaments\n"
    for limit in limits:
        print "%2i-limit"%et.primeNumbers[limit]
        write_lts(et.primes[:limit+1], nETs=nETs, etCutoff=0.2,
                highestComplexity=7, worstError=5e-4, func=func, error=error)

if __name__=='__main__':
    import sys
    if len(sys.argv)>1 and sys.argv[1].startswith('big'):
        bigTest()
    elif len(sys.argv)>1 and sys.argv[1].startswith('huge'):
        bigTest(nETs=1001)
    else:
        test()
