#  Regular temperament finding library
#  based on regular.py
#  only does weighted prime errors
#  and uses wedgies instead of the period mapping

import numpy

# evil generic import to ease backwards compatibility
from et import *

###############################################################################
#
#  Regular temperaments constructed from equal temperaments or mappings
#
###############################################################################


def defaultBadness(rt, width, error):
    m, n = rt.mapping[0,0], rt.mapping[1,0]
    rank = len(rt.mapping)
    return error * width**rank + (
        #fudge factor to give a unique ordering
        1e-10 * (1.618*m+n))

def getRegularTemperaments(ets, primes,
        highestComplexity,
        highestError,
        error=rmsError,
        badness=defaultBadness,
        useRMS=1,
        nResults=100,
        rank=2):
    """Returns a list of temperaments as specified,
    ordered by minimax badness
    """
    temperaments = TemperamentList()
    for mapping in combinations(rank, map(list, ets)):
        try:
            tuning = RegularTemperament(mapping, primes)
            if error in (rmsError, optimalRMSError):
                optimum = tuning.optimalRMSError()
            else:
                if useRMS:
                    tuning.optimizeRMS()
                else:
                    tuning.optimizeTOP()
                optimum = error(tuning.weightedPrimes())
            if optimum < highestError:
                width = tuning.wedgieComplexity()
                if width <= highestComplexity:
                    temperaments.add(tuning, badness(tuning, width, optimum))
        except DegenerateException:
            continue

    # check contorsion last
    # because it takes a long time to calculate
    # and we only need it for the good ones
    return [
        rt for rt in temperaments.retrieve()[:nResults+10]
        if rt.contorsion()==1][:nResults]

getLinearTemperaments = getRegularTemperaments

def Temperament(primes, *octaves):
    """return a Regular Temperament
       consistent with the given ETs
    """
    ets = [list(BestET(n, primes)) for n in octaves]
    return RegularTemperament(ets, primes) ###

temperament = Temperament # make a common typo legitimate


class RegularTemperament:
    def __init__(self, mapping, primes):

        self.primes = numpy.matrix(primes)
        self.mapping = numpy.matrix([list(el) for el in mapping])
        self.rank, self.dimension = self.mapping.shape

    def wedgie(self):
        if not hasattr(self, "wedgie_vector"):
            wedgie_vector = wedgie(self.mapping)
            try:
                if filter(None, wedgie_vector)[0] < 0:
                    wedgie_vector = tuple([-x for x in wedgie_vector])
            except IndexError:
                raise DegenerateException, "linearly dependent"
            self.wedgie_vector = wedgie_vector
        return self.wedgie_vector

    def invariant(self):
        """something that can be used as a unique key"""
        return self.wedgie()

    def wedgieComplexity(self):
        """Average absolute value of the weighted complexity"""
        # set the length so that the complexity of an
        # equal temperament is the number of notes to the octave
        return sum(map(abs, wedgie(self.weightedMapping()))
                )/self.dimension

    def maxWedgieComplexity(self):
        """Highest absolute value of the weighted complexity"""
        return max(map(abs, wedgie(self.weightedArray())))

    def newComplexity(self):
        """From the determinant of the mean-of-weighted-mapping matrix"""
        M = self.weightedMapping()
        meansMatrix = (M * M.transpose())/self.dimension
        return sqrt(numpy.linalg.det(meansMatrix))

    def optimalBadness(self):
        covarianceMatrix = numpy.cov(self.weightedMapping(), bias=1)
        return sqrt(numpy.linalg.det(covarianceMatrix))

    def scalarComplexity(self):
        """From the scalar product of the weighted wedgie"""
        # this is mathematically identical to the new complexity
        mapping = self.weightedArray()
        M = Multivector(mapping[0])
        for entry in mapping[1:]:
            M ^= Multivector(entry)
        return sqrt(abs((M^~M).flatten()[0])/ self.dimension**self.rank)

    def scalarBadness(self):
        def adjust(mapping):
            mean = sum(mapping)/self.dimension
            return [entry-mean for entry in mapping]
        mapping = self.weightedArray()
        M = Multivector(adjust(mapping[0]))
        for entry in mapping[1:]:
            M ^= Multivector(adjust(entry))
        return sqrt(abs((M^~M).flatten()[0])/self.dimension**self.rank)

    def scalarError(self):
        return self.scalarBadness()/self.scalarComplexity()

    def maxWedgieError(self):
        """One of Gene's ideas"""
        total = Multivector((1,)*self.dimension)
        for vec in self.weightedArray():
            total ^= Multivector(vec)
        return max(map(abs, total.flatten()))

    def optimalRMSError(self):
        """Return the optimal prime, RMS, weighted error"""
        residuals = self.leastSquaresResults()[1]
        try:
            return sqrt(residuals[0]/self.dimension)
        except IndexError:
            raise TemperamentException, "failed to optimize %s"%self.weightedMapping()

    def optimizeRMS(self):
        """Set the prime, RMS, weighted errors optimum"""
        self.basis = self.leastSquaresResults()[0]

    def leastSquaresResults(self):
        """Finds the RMS values using NumPy"""
        W = self.weightedMapping().transpose()
        P = numpy.ones(self.dimension)
        return numpy.linalg.lstsq(W, P, rcond=1e-20)

    def weightedPrimes(self):
        assert hasattr(self, "basis"), "Must have a tuning" 
        return numpy.array((self.basis * self.mapping)/self.primes)[0]

    def weightedMapping(self):
        """Optimization and complexity code uses this"""
        return self.mapping / self.primes

    def weightedArray(self):
        return numpy.array(self.weightedMapping())

    def optimizeTOP(self):
        """Set the TOP generators

        Requires PuLP and GLPK
        """
        import pulp
        prob = pulp.LpProblem("top", pulp.LpMinimize)

        # set some variables: the generators and the thing to minimize
        # all of them have to be positive
        generators = [
            pulp.LpVariable("generator%i"%n, 0, None)
            for n in range(self.rank)]
        error = pulp.LpVariable("error", 0, None)

        # specify that it's the error we want to minimize
        prob += error, "obj"

        # now set the errors of the temperament as constraints
        for map in self.weightedArray().transpose():
            weightedPrime = 0
            for generator, innermap in zip(generators, map):
                weightedPrime += generator*float(innermap)
            # set two error constraints for an overall absolute error
            prob += error >= weightedPrime - 1
            prob += error >= 1 - weightedPrime

        prob.solve(pulp.GLPK(msg=0))

        self.basis = numpy.array([g.varValue for g in generators])

    def contorsion(self):
        return hcf(self.wedgie())

    def __repr__(self, consonances=None):
        # send back useful stuff
        if not hasattr(self, 'basis'):
            self.optimizeRMS()
        w = self.weightedPrimes()
        params = {
            'wedgie' : self.wedgie(),
            'melody' : self.mapping,
            'tuning' : (self.basis * self.mapping)*1200,
            'scalarComplexity': self.newComplexity(),
            'rmsError' : rmsError(w)*1200,
            'maxError' : worstError(w)*1200,
            'steps' : self.basis*1200,
        }
        return """%(steps)s cent steps

wedgie:
%(wedgie)s

mapping by steps:
%(melody)s

tuning map (cents):
%(tuning)s

scalar complexity: %(scalarComplexity).3f
RMS weighted error: %(rmsError).3f cents/octave
max weighted error: %(maxError).3f cents/octave""" % params


class LinearTemperament(RegularTemperament):
    """Add some 2-specific things"""

    def complexity(self):
        """the range of complexities, a TOP-like measure"""
        oeMultivector = self.wedgie()[:self.dimension-1]
        oeMapping = numpy.array((0,) + oeMultivector)
        weightedMapping = oeMapping/self.primes[0,0]
        return max(weightedMapping)-min(weightedMapping)

    def __repr__(self, consonances=None):
        # send back useful stuff
        if not hasattr(self, 'basis'):
            self.optimizeRMS()
        w = self.weightedPrimes()
        params = {
            'wedgie' : self.wedgie(),
            'melody' : self.mapping,
            'tuning' : (self.basis * self.mapping)*1200,
            'complexity' : self.complexity(),
            'wedgieComplexity': self.wedgieComplexity(),
            'rmsError' : rmsError(w)*1200,
            'maxError' : worstError(w)*1200,
            'steps' : self.basis*1200,
        }

        return """%(steps)s cent steps

wedgie:
%(wedgie)s

mapping by steps:
%(melody)s

tuning map (cents):
%(tuning)s

octave-equivalent complexity: %(complexity).3f
wedgie based complexity: %(wedgieComplexity).3f
RMS weighted error: %(rmsError).3f cents/octave
max weighted error: %(maxError).3f cents/octave""" % params


class TemperamentList:
    def __init__(self):
        self.data = {}
    def add(self, temperament, badness):
        key = temperament.invariant()
        if key in self.data and badness > self.data[key][0]:
            return ###
        self.data[key] = badness, temperament
    def retrieve(self):
        decorated = self.data.values()
        decorated.sort()
        return [temperament for badness, temperament in decorated]


###############################################################################
#
#  General wedge products
#
###############################################################################

# These are simpler than the ones in temper.py
# they can't be constructed so many ways
# they can't be octave equivalent
# and they're all specific to a prime limit

class Multivector:
    def __init__(self, seed={}, size=None, basisSize=None):
        self.data = {}
        if size is None:
            # seed must be a sequence
            self.size = len(seed)
            assert basisSize is None or basisSize==1
            self.basisSize = 1
            for i in range(self.size):
                self[i,] = seed[i]
        else:
            self.size = size
            if hasattr(seed, 'items'):
                if basisSize is  None:
                    assert seed
                else:
                    self.basisSize = basisSize
                for key, value in seed.items():
                    if hasattr(self, "basisSize"):
                        assert self.basisSize == len(key)
                    else:
                        self.basisSize = len(key)
                    self[key] += value
            else:
                assert size == len(seed)
                assert basisSize is None or basisSize==1
                self.basisSize = 1
                for i in range(size):
                    self[i,] = seed[i]

    def invariant(self):
        """return the wedgie in vector form
        with the first non-zero entry positive
        """
        sign = None
        flat = self.flatten()
        nonZero = filter(None, flat)
        if (nonZero and nonZero[0]>0):
            return tuple(flat)
        return tuple([-x for x in flat])

    def flatten(self):
        seed = range(self.size)
        return [self[basis] for basis in combinations(self.basisSize, seed)]

    def torsion(self):
        return hcf(self.data.values())

    def __invert__(self):
        """The ~ operator does the complement"""

        result = Multivector(size=self.size, basisSize=self.size-self.basisSize)

        for key, value in self.data.items():
            newKey = [
                element for element in range(self.size)
                if element not in key]
            newKey = tuple(newKey)
            result[newKey] = equivalentValue(key+newKey, value)

        return result

    def __xor__(self, other):
        """The ^ operator does a wedge (exterior) product"""
        # some of the work here is done by __setitem__
        assert self.size == other.size
        newBasisSize = self.basisSize + other.basisSize
        assert newBasisSize <= self.size
        result = Multivector(size=self.size, basisSize=newBasisSize)
        for base1, value1 in self.data.items():
            for base2, value2 in other.data.items():
                value = value1 * value2
                for element in base1:
                    if element in base2:
                        break
                else:
                    base = base1 + base2
                    result[base] += value
        return result

    def __getitem__(self, key):
        base, value = self.wedgeEquivalent(key, 1)
        return self.data.get(base, 0)*value

    def __setitem__(self, key, value):
        base, value = self.wedgeEquivalent(key, value)
        if value:
            self.data[base] = value
        else:
            if self.data.has_key(base):
                del self.data[base]

    def wedgeEquivalent(self, base, value):
        """convert the base and value into a
           simpler form consistent with the
           rules of exterior products
        """
        if not isinstance(base, tuple):
            raise KeyError, "Multivector keys must be tuples"
        if len(base) != self.basisSize:
            raise KeyError, "Key is the wrong size for this wedgie"
        value = equivalentValue(base, value)
        baseCopy = list(base)
        baseCopy.sort()
        return tuple(baseCopy), value

    def __repr__(self):
        return '('*self.basisSize + str(self.flatten())[1:-1] + ')'*self.basisSize


def equivalentValue(base, value):
    for i in range(len(base)-1):
        for j in range(i+1, len(base)):
            if base[i] == base[j]:
                return 0 ###
            if base[j]<base[i]:
                value = -value
    return value

def combinations(number, input):
    """Return all combinations of number from the input list."""
    if number==1:
        return [(next,) for next in input]
    assert number>1
    output = []
    for i in range(len(input)-number+1):
        for element in combinations(number-1, input[i+1:]):
            next = (input[i],) + element
            output.append(next)
    return output

def wedgie(mapping):
    """Return the wedgie for the linear temperament with these
       ET mappings, in vector form.
    """
    mapping = numpy.array(mapping)
    total = Multivector(mapping[0])
    for entry in mapping[1:]:
        total ^= Multivector(entry)
    return tuple(total.flatten())


###############################################################################
#
#  Test code
#
###############################################################################

# originally duplicated the output of write_temper.ml
# now based on write_temper_rms.py
# only does RMS stuff because TOP optimization is so slow

def write_rts(primes, etCutoff=0.17,
        nETs=20, highestComplexity=2, worstError=0.002,
        useRMS=1, error=rmsError, rank=2):
    # make adjustments
    highestComplexity *= len(primes)**2 * 0.24
    etCutoff /= sqrt(len(primes))

    ets = getEqualTemperaments(primes, nETs=nETs, cutoff=etCutoff, error=error)
    print ' '.join([str(et[0]) for et in ets])

    for lt in getRegularTemperaments(ets, primes, highestComplexity, worstError,
            rank=rank, useRMS=useRMS, error=error):
        print lt.invariant()
    print "\n"

write_lts = write_rts

tubulongFrequencies = numpy.array([
    float(s*(s*s - 1)) / sqrt((s*s + 1) * 7.2)
    for s in range(3,9)])

tubulongPrimes = numpy.log(tubulongFrequencies)/log(2)

def test():
    limits = range(2, 11) # indexes primes (or primeNumbers)
    print "RMS Optima\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_rts(primes[:limit+1])

    print "Less good ETs\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_rts(primes[:limit+1], etCutoff=0.22, worstError=0.005)

    print "Microtemperaments\n"
    for limit in limits:
        print "%2i-limit" % primeNumbers[limit]
        write_rts(primes[:limit+1], nETs=30, highestComplexity=5, worstError=5e-4)

    for f in tubulongFrequencies:
      print "%.5f"%f,
    print ''

    print "Octave-Equivalent Tubulongs\n"
    oeTubulongPrimes = (1.0,)+tuple(tubulongPrimes)
    for limit in range(2, len(oeTubulongPrimes)):
        write_rts(oeTubulongPrimes[:limit+1], worstError = 0.002)

    print "Straight Tubulongs\n"
    for limit in range(2, len(tubulongPrimes)):
        write_rts(tubulongPrimes[:limit+1], worstError = 0.002)

if __name__=='__main__':
    test()
