"""Various utilities for temperament searches"""

from __future__ import division

import math

#
# Defining the prime intervals
#

# Intervals are measured in octaves.
# In general, the first entry in a list of primes is taken to be
# the equivalence interval.
# Tenney weighting is implied.
def log2(f):
    return math.log(f)/math.log(2)
primeNumbers = 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31
primes = [log2(f) for f in primeNumbers]

def makeLimit(limit):
    """Returns the labels and primes for the limit given as a list.
    If it's a list of one number, that specifies the prime limit.
    If more, they specify the prime intervals.
    An integer means a harmonic.
    A pair of integers means a rational partial.
    A float means a partial specified in cents.
    The labels are strings when they aren't integers.
    All kinds of exceptions may be thrown for invalid input.
    """
    if len(limit)==1:
        try:
            highest, = limit[0]
        except TypeError:
            highest, = limit
        for i in range(len(primeNumbers)):
            if primeNumbers[i] > highest:
                return primeNumbers[:i], primes[:i]
        return primeNumbers, primes
    else:
        labels, plimit = [], []
        for partial in limit:
            if isinstance(partial, int):
                labels.append(partial)
                plimit.append(log2(partial))
            elif isinstance(partial, float):
                labels.append('%.3f'%partial)
                plimit.append(partial/12e2)
            else:
                if len(partial)==1:
                    # assume this should be a harmonic
                    partial = partial[0], 1
                n, d = partial
                if d==1:
                    labels.append(n)
                    plimit.append(log2(n))
                else:
                    labels.append('%i/%i' % (n, d))
                    plimit.append(log2(n) - log2(d))
        return labels, plimit

def arePrimeNumbers(numbers):
    return tuple(numbers) == primeNumbers[:len(numbers)]

def formatPrimeLabels(labels, separator='.'):
    if arePrimeNumbers(labels):
        return str(labels[-1])
    if separator == '.':
        for label in labels:
            if isinstance(label, str) and '.' in label:
                separator = ', '
                break
    return separator.join(map(str, labels))

standardWarts = dict(zip(primeNumbers, "abcdefghijklmn"))
otherWarts = "qrstuvwxyzabcdefghijklmnop"
def wartsFromPrimeLabels(labels):
    """Return some letters to be used in suffixes for a
    variant of Herman's ET naming scheme.
    Letters from the start of the alphabet are used for primes.
    Letters starting from q are used otherwise.
    """
    wartStore = iter(otherWarts)
    result = []
    for label in labels:
        if label in standardWarts:
            result.append(standardWarts[label])
        else:
            result.append(wartStore.next())
    return ''.join(result)

def textToLimit(text):
    partials = []
    if '_' in text:
        # written by CGI
        for partial in text.split('_'):
            if '.' in partial:
                partials.append(float(partial))
            else:
                partials.append(map(int, partial.split('/')))
    elif ' ' in text:
        # a list of values in cents
        for partial in text.split():
            if partial.endswith(','):
                partial = partial[:-1]
            partials.append(float(partial))
    elif ':' in text:
        # partials specified as a chord
        freqs = text.split(':')
        root = int(freqs[0])
        for partial in freqs[1:]:
            if '.' in partial:
                # convert to cents
                partials.append(log2(float(partial)/root)*1200)
            elif '/' in partial:
                n, d = map(int, partial.split('/'))
                d *= root
                if n < d:
                    d, n = n, d
                partials.append((n, d)) # don't bother to reduce
            else:
                n, d = int(partial), root
                if n < d:
                    d, n = n, d
                partials.append((n, d)) # don't bother to reduce
    else:
        # . separated harmonics or ratios
        for partial in text.split('.'):
            if '/' in partial:
                n, d = map(int, partial.split('/'))
                if n < d:
                    d, n = n, d
                partials.append((n, d)) # don't bother to reduce
            else:
                partials.append(int(partial))
    return makeLimit(partials)

#
# The TemperamentList class
#

class TemperamentList:
    """Used for the containers temperaments are stored in during searches.
    Stores up to the given number of Rank2Temperaments ordered by badness.
    This is optimized for efficiency of large searches.
    It can reject temperaments that are too bad to make the list.
    It can also reject temperaments that are no better than equivalent
    ones it's seen before.
    """
    def __init__(self, capacity):
        self.capacity = capacity
        self.data = {} # list of the things we've seen before
        self.queue = [] # the best ETs sorted by badness

    def goodEnough(self, badness):
        """Say of a temperament with this badness would make the list"""
        return len(self.queue)<self.capacity or badness<self.queue[-1][0]

    def add(self, temperament, badness):
        """Add a temperament to the list, if it's good and novel enough.
        Trust the caller to supply the correct badness to save
        re-calculating it.
        """
        key = temperament.invariant()
        newItem = badness, temperament
        if self.data.has_key(key):
            oldItem = self.data[key]
            if newItem >= oldItem:
                return ###
            try:
                self.queue.remove(oldItem)
            except ValueError:
                pass
        self.data[key] = newItem
        self.queue.append(newItem)
        self.queue.sort()
        del self.queue[self.capacity:]

    def retrieve(self):
        """Return all the stored temperaments"""
        return [lt for badness, lt in self.queue]

class ConservativeList(TemperamentList):
    """Always keeps the oldest"""
    def add(self, temperament, badness):
        """Add a temperament to the list, if it's good and novel enough.
        Trust the caller to supply the correct badness to save
        re-calculating it.
        """
        key = temperament.invariant()
        if key not in self.data:
            newItem = badness, temperament
            self.data[key] = newItem
            self.queue.append(newItem)
            self.queue.sort()
            del self.queue[self.capacity:]

#
# Statistical functions
#

def mean(x):
    """The arithmetic mean of a sequence"""
    return sum(x)/len(x)

def rms(x):
    """The root mean squared of a sequence"""
    return math.sqrt(mean([y*y for y in x]))

def lae(x, y):
    """x is less then or approximately equal to y"""
    return x <= y or 0 < x/y < 1.0000001
#
# invariants as reduced lattices
#

def nameThatMapping(mapping, labels=None):
    """Returns a pair of the name of a temperament class
    and the number of primes you need to remove from the mapping
    to get it.
    Raises a KeyError if no name can be found.
    """
    import names
    r = len(mapping)
    dimension = len(mapping[0])
    if __debug__:
        for row in mapping:
            assert len(row) == dimension
    if labels is not None:
        assert len(labels) == dimension
        limit = tuple(labels)
        if limit in names.namesByLimit:
            key = lattice_invariant(mapping)
            if key in names.namesByLimit[limit]:
                # An exact non-canonical name takes priority
                return names.namesByLimit[limit][key], 0
        # check the canonical order next
        newMapping, newLimit = canonical_order(mapping, limit)
        if newLimit != limit:
            return nameThatMapping(newMapping, newLimit)
    for ignoredDimensions in range(dimension-2) or [0]:
        if ignoredDimensions == 0:
            submapping = mapping
        else:
            submapping = [row[:-ignoredDimensions] for row in mapping]
        if rank(submapping) < r:
            r = rank(submapping)
        invariant = lattice_invariant(submapping, r)
        if labels is None:
            lookup = names.namesByRank.get(r, {})
        else:
            lookup = names.namesByLimit.get(tuple(labels[:-ignoredDimensions]), {})
        if invariant in lookup:
            if not (ignoredDimensions and r==(dimension-ignoredDimensions)): # JI
                return lookup[invariant], ignoredDimensions
    if labels is None:
        raise KeyError, "no name found"
    name, ignored = nameThatMapping(mapping)
    return name+'?', ignored

def canonical_order(mapping, labels):
    decorated = zip(labels, *mapping)
    decorated.sort()
    transpose = zip(*decorated)
    return transpose[1:], transpose[0]

def lattice_invariant(mapping, rank=None):
    if rank is None:
        rank = len(mapping)
    reduced = lattice_reduction(mapping)
    for entry in reduced[rank:]:
        assert len(filter(None, entry)) == 0
    del reduced[rank:]
    for i in range(len(reduced)):
        assert len(filter(None, reduced[i][:i])) == 0
        reduced[i] = reduced[i][i:]
    reduced.reverse()
    result = []
    for entry in reduced:
        result.extend(entry)
    return tuple(result)

def mappingFromInvariant(invariant, rank):
    length = len(invariant)
    mapping = []
    remainder = tuple(invariant)
    padding = rank - 1
    total_padding = (padding*(padding+1))>>1
    n_dimensions, check = divmod(len(invariant) + total_padding, rank)
    if check:
        raise IndexError, "Invariant of wrong length for rank"
    for r in range(rank):
        mapping.append((0,)*padding + remainder[:(n_dimensions-padding)])
        remainder = remainder[(n_dimensions-padding):]
        padding -= 1
    mapping.reverse()
    return mapping

def lattice_reduction(ets):
    """Reduced echelon form without removing common factors
    and allowing positive numbers to appear in place of zeros
    instead of introducing torsion.
    (These positive numbers should be the smallest possible.)
    """
    echelon = echelon_form(ets)
    n_columns = len(echelon[0])
    column = 1
    for row in range(1, len(ets)):
        try:
            while echelon[row][column]==0:
                column += 1
        except IndexError:
            # lower rank
            break
        n = echelon[row][column]
        for i in range(row):
            m = divmod(echelon[i][column], n)[0]
            for j in range(n_columns):
                echelon[i][j] -= m*echelon[row][j]
    return echelon

def rank(mapping):
    result = 0
    for row in echelon_form(mapping):
        for element in row:
            if element:
                break
        else:
            return result
        result += 1
    return result

def echelon_form(ets):
    """Echelon form without removing common factors."""
    rows = [list(et) for et in ets]
    n_rows = len(rows)
    n_columns = len(rows[0])
    for row in rows:
        assert len(row)==n_columns
    column = 0
    for row in range(n_rows-1):
        sub_matrix = [normalize_sign(r) for r in rows[row:]]
        sub_row = 0
        while sub_row < len(sub_matrix) - 1:
            sub_matrix.sort()
            while column < n_columns:
                try:
                    while sub_matrix[sub_row][column] == 0:
                        sub_row += 1
                    break
                except IndexError:
                    column += 1
                    sub_row = 0
            else:
                # all zeros from here
                break
            for i in range(sub_row+1, len(sub_matrix)):
                n = divmod(sub_matrix[i][column],
                           sub_matrix[sub_row][column])[0]
                for j in range(column, n_columns):
                    sub_matrix[i][j] -= n*sub_matrix[sub_row][j]
        sub_matrix.reverse()
        rows[row:] = sub_matrix
    rows[-1] = normalize_sign(rows[-1])
    return rows

def standard_form(vectors, plimit):
    """TLLL-reduced mapping"""
    buoys = [p*p for p in plimit]
    def tenneyEuclid(u, v):
        assert len(u)==len(v) and len(u) <= len(buoys)
        return sum([x*y/w for x, y, w in zip(u, v, buoys)])
    return LLL(vectors, tenneyEuclid)

def LLL(vectors, prod):
    """LLL reduction, with a suitable Euclidian inner product prod"""
    # from von zur Gathen & Gerhard p. 452
    g, n = [list(row) for row in vectors], len(vectors)
    gs, m = GSO(g, prod)
    i = 1
    while i < n:
        for j in reversed(range(i)):
            g[i] = [x - roundLLL(m[i][j]) * y for (x, y) in zip(g[i], g[j])]
            gs, m = GSO(g, prod) # indentation?
        if i > 0 and prod(gs[i-1], gs[i-1]) > 2*prod(gs[i], gs[i]):
            g[i-1], g[i] = g[i], g[i-1]
            gs, m = GSO(g, prod)
            i -= 1
        else:
            i += 1
    return g

def roundLLL(x):
    """The special rounding function that LLL reduction needs
    where 0.5 rounds down.
    c.f. von zur Gathen & Gerhard, p.458
    and the absence of a final rep(3,2)
    """
    return int(round(x - 1e-10))

def GSO(basis, prod):
    """
    Gram-Schmidt orthogonalization in the inner product space of prod.
    Returns the Gramian matrix and the other thing as a pair
    """
    # from von zur Gathen & Gerhard p. 449
    m, gramian, idrow = [], [], [1] + [0]*len(basis)
    for frow in basis:
        del idrow[-1]
        row = [prod(frow, g) / prod(g, g) for g in gramian] + idrow
        for mcell, grow in zip(row, gramian):
            frow = [x - mcell*g for (x, g) in zip(frow, grow)]
        gramian.append(frow)
        m.append(row)
    return gramian, m

def real_echelon_reduce(matrix):
    """Echelon form of a real (float) matrix in place"""
    size = len(matrix)
    if size < 2:
        return
    width = len(matrix[0])
    for row in matrix[1:]:
        assert width == len(row)
    first_row = col = 0
    for col in range(width):
        # Reduce the ith column from the ith row
        for row in range(first_row, size):
            if matrix[row][col] != 0.0:
                break
        else:
            # column all zeros
            continue
        # swap rows
        matrix[first_row], matrix[row] = matrix[row], matrix[first_row]
        # zero the rest of the column
        for row in range(first_row+1, size):
            mul = float(matrix[row][col])/matrix[first_row][col]
            for sub_col in range(col, width):
                matrix[row][sub_col] -= mul * matrix[first_row][sub_col]
        first_row += 1
        if first_row == size - 1:
            return

def normalize_sign(mapping):
    nonzeros = [m for m in mapping if m]
    if nonzeros==[] or nonzeros[0]>0:
        return mapping
    return [-m for m in mapping]

def size_of_matrix(m):
    size = len(m)
    n_primes = len(m[0])
    assert size <= n_primes
    ms_rows = []
    for row in m:
        assert len(row) == n_primes
        ms_rows.append(mean([x*x for x in row]))
    if size==1:
        result = ms_rows[0]
    elif size==2:
        cross_term = mean([x*y for x, y in zip(m[0], m[1])])
        result = ms_rows[0]*ms_rows[1] - cross_term**2
    else:
        mtm = [[0]*size for i in range(size)]
        for i in range(size):
            mtm[i][i] = ms_rows[i]
            for j in range(i+1, size):
                mtm[i][j] = mtm[j][i] = mean([x*y for x, y in zip(m[i], m[j])])
        real_echelon_reduce(mtm)
        result = 1.0
        for i in range(size):
            result *= mtm[i][i]
    if result < 0:
        return 0.0
    return math.sqrt(result)

#
# Interval vectors and mappings
#

def tempersOut(mapping, vector):
    for element in vector[len(mapping):]:
        if element != 0:
            return False
    return sum([m*v for m, v in zip(mapping, vector)])==0

def formatMapping(mapping):
    return '<' + repr(list(mapping))[1:]

def formatInterval(interval):
    return repr(list(interval))[:-1] + '>'

def formatMappingMatrix(m):
    return '[%s>' % ',\n '.join(map(formatMapping, m))

def inherentError(uv, plimit):
    """Returns the TOP-RMS error for the best case of
    tempering out the given unison vector in the given
    prime limit.
    The number of dimensions does matter.
    """
    # See http://x31eq.com/primerr.pdf p.22 equation 107
    q = [x*p for x, p in zip(fitToPrimeLimit(uv, plimit), plimit)]
    return abs(mean(q))/rms(q)

def fitToPrimeLimit(uv, plimit):
    d = len(plimit)
    assert not filter(None, uv[d:])
    return (list(uv) + [0]*(d - len(uv)))[:d]

def setBestLimit(intervals):
    """Returns the smallest consecutive prime limit
    that works with the given list of intervals
    and also returns those intervals all of the right dimension.
    Assumes harmonic timbres.
    """
    size = max(map(last_non_zero, intervals)) + 1
    result = []
    for interval in intervals:
        if len(interval) >= size:
            result.append(tuple(interval[:size]))
        else:
            result.append(tuple(interval) + (0,)*(size - len(interval)))
    return result, primes[:size]

def last_non_zero(interval):
    result = 0
    for i in range(len(interval)):
        if interval[i] != 0:
            result = i
    return result

#
# Ratios
#

ratioPattern = '(\d+)[/:](\d+)'
vectorPattern = '[[|(]\s*(-?\d+(?:[, ]\s*-?\d+)*)\s*[])>]'

def factorizeRatio(n, d, labels=primeNumbers):
    """return n/d as a ratio space vector"""
    assert n > 0 and d > 0
    return tuple([n-d for n, d in
        zip(factorize(n, labels), factorize(d, labels))])

def factorize(n, labels=primeNumbers):
    if __debug__:
        for num in primeNumbers:
            assert type(num) == int
    assert n>0
    result = []
    residue = n
    for prime in labels:
        for factors in range(1000): # make the range bigger for stupid ratios
            newResidue, remainder = divmod(residue, prime)
            if remainder != 0:
                break
            residue = newResidue
        else:
            raise OverflowError("Number too large to factorize")
        result.append(factors)
    if residue != 1:
        raise OverflowError("Inappropriate prime limit")
    return tuple(result)

def getRatio(interval, primes=primeNumbers):
    """convert the octave-specific ratio space vector
    to a ratio n/d

    return (n,d)

    """
    n = d = 1
    for r, partial in zip(interval, primes):
        if r == 0:
            continue
        if isinstance(partial, int):
            nx, dx = partial, 1
        elif isinstance(partial, str):
            nx, dx = map(int, partial.split("/"))
        else:
            raise ValueError, "Bad basis for rationalization"
        if r > 0:
            n *= nx**r
            d *= dx**r
        else:
            n *= dx**(-r)
            d *= nx**(-r)
    assert type(n) in (int, long), (interval, primes, n)
    assert type(d) in (int, long), (interval, primes, d)
    return n, d

def textToIntervals(s, labels=primeNumbers):
    """Used by the CGI"""
    import re
    vectors = []
    for n, d in re.findall(ratioPattern, s):
        try:
            vectors.append(factorizeRatio(int(n), int(d), labels))
        except TypeError:
            raise TypeError, "You can only factorize ratios with integer harmonics"
    for result in re.findall(vectorPattern, s):
        numbers = map(int, result.replace(',', ' ').split())
        vectors.append(tuple(numbers + [0]*(len(labels)-len(numbers))))
    return vectors

def dotprod(u, v):
    assert len(u) == len(v)
    return sum([x*y for (x, y) in zip(u, v)])

if __name__=='__main__':

    # Example from von zur Gathen & Gerhard p.458
    # that happens to use the same input matrix as Wikipedia
    # but gives a different result.
    ggmat = [[1,1,1],[-1,0,2],[3,5,6]]
    assert LLL(ggmat, dotprod) == [[0,1,0], [1,0,1], [-1, 0, 2]]
    # Example from von zur Gathen & Gerhard p.453
    assert LLL([[12,2],[13,4]], dotprod) == [[1, 2], [9, -4]]

