#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include "temper.h"

/*
 * C code for temperament finding library
 *
 * see http://www.microtonal.co.uk/temper/
 *
 */

/*
 * general utilities
 */

double magnitude(float_array* primes, int_array* interval) {
    assert(primes != NULL);
    assert(interval != NULL);
    double total;
    int first_prime;
    int size = primes->size;
    if (interval->size > size) {
        assert(interval->size == (size+1));
        total = (double) interval->data[0];
        first_prime = 1;
    }
    else {
        total = 0.0;
        first_prime = 0;
    }
    for (int i=0; i<size; i++) {
        total += ((double) interval->data[i+first_prime]) * primes->data[i];
    }
    return total;
}

long nint(double x) {
    if (x>0.0) {
        return (long) (x+0.5);
    }
    else {
        return -((long) (0.5-x));
    }
}

long gcd(long m, long n) {
    /* Euclid's algorithm */
    while (n!=0) {
        long swap;
        swap = n;
        n = m%n;
        m = swap;
    }
    return labs(m);
}

void extended_euclid(
    long last_t, long last_s, long last_r,
    long next_t, long next_s, long next_r,
    long *return_t, long *return_s, long *return_r) {

    /*
     * Extended Euclidian algorithm
     * from von zur Gathen & Gerhard, p.47
     */

    while (next_r!=0) {
        long q, r, sign_of_r;
        long swap_t, swap_s, swap_r;

        q = last_r / next_r;
        r = last_r % next_r;
        sign_of_r = r<0? -1: 1;

        swap_t = next_t;
        swap_s = next_s;
        swap_r = next_r;

        next_t = (last_t - q*next_t) / sign_of_r;
        next_s = (last_s - q*next_s) / sign_of_r;
        next_r = abs(r);

        last_t = swap_t;
        last_s = swap_s;
        last_r = swap_r;
    }
    *return_t = last_t;
    *return_s = last_s;
    *return_r = last_r;
}

/*
 * equal temperaments
 */

equal_temperament* equal_temperament_list_head(generic_list *this) {
    return (equal_temperament*) generic_list_head(this);
}

equal_temperament* equal_temperament_new(int_array* mapping) {
    assert(mapping != NULL);
    equal_temperament* this = (equal_temperament*)
            malloc(sizeof(equal_temperament));
    if (this != NULL) {
        this -> mapping = mapping;
    }
    return this;
}

int_array* equal_temperament_mapping(equal_temperament* this) {
    assert(this != NULL);
    return this->mapping;
}

float_array* equal_temperament_errors(
        equal_temperament* this,
        float_array* primes) {
    assert(this != NULL);
    assert(primes != NULL);
    float_array* errors = float_array_new(primes->size);
    if (  primes->size == (this->mapping->size - 1) &&
            primes->size > 0) {
        double octave = (double) this->mapping->data[0];
        for (int i=0; i<primes->size; i++) {
            errors->data[i] = (
                ((double) this->mapping->data[i+1]) -
                primes->data[i] * octave);
        }
    }
    return errors;
}

double equal_temperament_resolve(
        equal_temperament* this,
        int_array* interval) {
    assert(this != NULL);
    assert(interval != NULL);
    int size = interval->size;
    int first_prime = 0;
    double total = 0.0;
    if (size < this->mapping->size) {
        first_prime = 1;
    }
    for (int i=0; i<size; i++) {
        total += (double) this->mapping->data[i+first_prime]
            * interval->data[i];
    }
    return total / ((double) this->mapping->data[0]);
}

double equal_temperament_worst_error(
        equal_temperament* this,
        generic_list* intervals,
        float_array* primes) {
    assert(this != NULL);
    assert(intervals != NULL);
    assert(primes != NULL);
    int_array* interval;
    double error, ideal_size, tempered_size;

    double worst_error = 0.0;
    generic_list* node = intervals;
    while (node != NULL) {
        interval = (int_array*) generic_list_head(node);
        assert(interval != NULL);
        ideal_size = magnitude(primes, interval);
        tempered_size = equal_temperament_resolve(this, interval);
        error = fabs(ideal_size-tempered_size);
        if (error>worst_error) {
            worst_error = error;
        }
        node = generic_list_tail(node);
    }
    return worst_error * ((double) this->mapping->data[0]);
}

void equal_temperament_free(equal_temperament* this) {
    if (this != NULL) {
        free(this);
        /* somebody else has to free the mapping */
    }
}

generic_list* alternative_mappings_rec(
        equal_temperament* et,
        float_array* primes,
        double cutoff,
        int index,
        generic_list* running_total) {
    /* recursive function for alternative_mappings below
     */

    assert(et != NULL);
    assert(primes != NULL);
    assert(index>=0);

    generic_list* result = running_total;
    int size = et->mapping->size;
    assert(index <= size);
    if (index == size) {
        result = generic_list_add(result, et);
    }
    else {
        double target, approximation, error;
        equal_temperament* new_et;
        int_array* basis;
        double octave = (double) et->mapping->data[0];

        result = alternative_mappings_rec
            (et, primes, cutoff, index+1, result);

        target = primes->data[index-1] * octave;
        approximation = (double) et->mapping->data[index];
        error = approximation - target;
        if (1.0 - fabs(error) < cutoff) {
            basis = int_array_new(size);
            for (int j=0; j<size; j++) {
                basis->data[j] = et->mapping->data[j];
            }
            new_et = equal_temperament_new(basis);
            assert (new_et != NULL);

            basis->data[index] += (error<0.0)? 1: -1;
            result = alternative_mappings_rec
                (new_et, primes, cutoff, index+1, result);
        }
    }
    return result;
}

generic_list* alternative_mappings(
        equal_temperament* et,
        float_array* primes,
        double cutoff) {
    /* Get all equal temperaments that might be consistent
     * I don't think cutoffs higher than 1.0 are meaningful
     */
    /* copy the input so that it doesn't get garbage collected twice */
    int_array* basis = int_array_new(et->mapping->size);
    for (int j=0; j<et->mapping->size; j++) {
        basis->data[j] = et->mapping->data[j];
    }
    equal_temperament* new_et = equal_temperament_new(basis);
    assert (new_et != NULL);
    return alternative_mappings_rec(
        new_et, primes, cutoff, 1, generic_list_new());
}

equal_temperament* prime_et(long n_notes, float_array* primes) {
    assert(n_notes>0);
    assert(primes != NULL);
    int size = primes->size + 1;
    int_array* mapping = int_array_new(size);
    if (mapping != NULL) {
        mapping->data[0] = n_notes;
        for(int i=1; i<size; i++) {
            mapping->data[i] = nint(
                ((double) n_notes) * primes->data[i-1]);
        }
        return equal_temperament_new(mapping); /***/
    }
    return NULL;
}

equal_temperament* best_et(
        long n_notes,
        generic_list* consonances,
        float_array* primes) {

    assert(consonances != NULL);
    assert(primes != NULL);

    equal_temperament* seed_et = prime_et(n_notes, primes);
    generic_list* ets = alternative_mappings(seed_et, primes, 1.0);
    int_array_free(seed_et->mapping);
    equal_temperament_free(seed_et);

    generic_list* remainder = ets;
    double minimax_error = 1e10;
    equal_temperament* et;
    equal_temperament* best_et = NULL;
    double error;
    while (remainder != NULL) {
        et = (equal_temperament*) generic_list_head(remainder);
        error = equal_temperament_worst_error(et, consonances, primes);
        if (error<minimax_error) {
            minimax_error = error;
            best_et = et;
        }
        remainder = generic_list_tail(remainder);
    }

    /* now clean up the garbage */
    remainder = ets;
    while (remainder != NULL) {
        et = (equal_temperament*) generic_list_head(remainder);
        if (et != best_et) {
            int_array_free(et->mapping);
            equal_temperament_free(et);
        }
        remainder = generic_list_tail(remainder);
    }
    generic_list_free(ets);

    return best_et;
}

generic_list* get_limited_ets(
        generic_list* consonances,
        float_array* primes,
        int n_ets,
        double cutoff) {
    assert(primes != NULL);
    assert(n_ets >= 0);

    generic_list* result = generic_list_new();
    int ets_found = 0;
    long n_notes = 1;
    equal_temperament* et;
    generic_list* candidates;
    generic_list* remainder;
    long running_gcd;

    while(ets_found < n_ets) {
        et = prime_et(n_notes, primes);
        candidates = generic_list_new();
        if (cutoff <= 0.5) {
            candidates = generic_list_add(candidates, et);
        }
        else {
            candidates = alternative_mappings(et, primes, cutoff);
            int_array_free(et->mapping);
            equal_temperament_free(et);
        }
        remainder = candidates;
        while (remainder != NULL) {
            et = (equal_temperament*) generic_list_head(remainder);
            assert(et != NULL);
            running_gcd = 0;
            for(int i=0; i<et->mapping->size; i++) {
                running_gcd = gcd(running_gcd, et->mapping->data[i]);
            }
            if (
                    running_gcd==1 &&
                    equal_temperament_worst_error(
                        et, consonances, primes) < cutoff) {
                result = generic_list_add(result, et);
                ets_found++;
            }
            else {
                int_array_free(et->mapping);
                equal_temperament_free(et);
            }
            remainder = generic_list_tail(remainder);
        }
        generic_list_free(candidates);

        n_notes++;
    }
    return result;
}


/*
 * Linear Mappings
 */

linear_mapping* linear_mapping_list_head(generic_list *this) {
    return (linear_mapping*) generic_list_head(this);
}

linear_mapping* linear_mapping_new(int_array* map1, int_array* map2) {
    assert(map1 != NULL);
    assert(map2 != NULL);
    assert(map1->size == map2->size);

    linear_mapping* this = (linear_mapping*)
            malloc(sizeof(linear_mapping));
    if (this != NULL) {
        int size, i;
        long m, n;
        long m_gen, n_gen, octave_division;
        this -> et1 = map1;
        this -> et2 = map2;
        size = map1->size;

        m = map1->data[0];
        n = map2->data[0];
        assert(m>=0);
        assert(n>=0);
        extended_euclid(0, 1, m, -1, 0, n,
            &m_gen, &n_gen, &octave_division);

        /* get primes in terms of equivalence interval and generator */
        this->period_mapping = int_array_new(size);
        assert(this->period_mapping != NULL);
        this->generator_mapping = int_array_new(size-1);
        assert(this->generator_mapping != NULL);
        assert(labs(m*n_gen - n*m_gen) == octave_division);
        assert(octave_division>0);
        this->period_mapping->data[0] = octave_division;
        for (i=1; i<size; i++) {
            this->period_mapping->data[i] =
                n_gen*map1->data[i] - m_gen*map2->data[i];
            this->generator_mapping->data[i-1] =
                (m*map2->data[i] - n*map1->data[i])
                    / octave_division * (m_gen<0? -1:1);
        }
    }
    return this;
}

int_array* linear_mapping_et1(linear_mapping* this) {
    assert(this != NULL);
    return this->et1;
}

int_array* linear_mapping_et2(linear_mapping* this) {
    assert(this != NULL);
    return this->et2;
}

int_array* linear_mapping_period_mapping(linear_mapping* this) {
    assert(this != NULL);
    return this->period_mapping;
}

int_array* linear_mapping_generator_mapping(linear_mapping* this) {
    assert(this != NULL);
    return this->generator_mapping;
}

int_array* linear_mapping_invariant(linear_mapping* this) {
    /* return an array that uniquely identifies the mapping */
    /* the receiver has to free! */
    assert(this != NULL);
    int_array* result = int_array_new(this->period_mapping->size);
    if (result != NULL) {
        int i;
        long sign=0;
        result->data[0] = this->period_mapping->data[0];
        for (i=0; i<this->generator_mapping->size; i++) {
            if (this->generator_mapping->data[i] != 0) {
                sign = this->generator_mapping->data[i] < 0? -1: 1;
                break;
            }
        }
        assert(sign != 0);
        for (i=0; i<this->generator_mapping->size; i++) {
            result->data[i+1] = this->generator_mapping->data[i] * sign;
        }
    }
    return result;
}

long linear_mapping_width(linear_mapping* this, int_array* interval) {
    int i;
    long total=0;
    assert(this != NULL);
    assert(interval != NULL);
    assert(this->period_mapping->size == interval->size);

    for (i=0; i<this->generator_mapping->size; i++) {
        total += this->generator_mapping->data[i] * interval->data[i+1];
    }
    return labs(total);
}

int_array* linear_mapping_widest
        (linear_mapping* this, generic_list* intervals) {
    generic_list* remainder;
    int_array* widest = NULL;
    long max_width = 0;
    int_array* current;
    assert(this != NULL);
    remainder = intervals;

    while (remainder != NULL) {
        current = (int_array*) generic_list_head(remainder);
        assert(current != NULL);
        if (linear_mapping_width(this, current) > max_width) {
            widest = current;
            max_width = linear_mapping_width(this, current);
        }
        remainder = generic_list_tail(remainder);
    }

    return widest;
}

long linear_mapping_max_width(linear_mapping* this, generic_list* intervals) {
    return linear_mapping_width(this, linear_mapping_widest(this, intervals));
}

void linear_mapping_free(linear_mapping* this) {
    if (this != NULL) {
        int_array_free(this->period_mapping);
        int_array_free(this->generator_mapping);
        free(this);
        /* somebody else has to free the ET mappings */
    }
}

/*
 * Linear Temperaments
 * (linear mappings with a tuning)
 */

linear_temperament* linear_temperament_new(
        linear_mapping* mapping, double generator_size) {

    assert(mapping != NULL);

    linear_temperament *this = (linear_temperament *)
            malloc(sizeof(linear_temperament));
    if (this != NULL) {
        this->mapping = mapping;
        this->period_size = 1.0 / (double)mapping->period_mapping->data[0];
        this->generator_size = generator_size;
    }
    return this;
}

void linear_temperament_free(linear_temperament* this) {
    free(this);
}

linear_mapping* linear_temperament_as_linear_mapping(
        linear_temperament* this) {
    return this->mapping;
}

float_array* linear_temperament_errors(
        linear_temperament* this,
        float_array* primes) {

    assert(this != NULL);
    assert(primes != NULL);
    assert(primes->size > 0);
    int_array* period_mapping = this->mapping->period_mapping;
    int_array* generator_mapping = this->mapping->generator_mapping;
    float_array* errors = float_array_new(primes->size);
    assert(period_mapping != NULL);
    assert(generator_mapping != NULL);
    assert(errors != NULL);
    int offset;
    if (primes->size == period_mapping->size) {
        offset = 1;
        double tempered_size = (double) int_array_get(period_mapping,0)
                * this->period_size;
        float_array_set(errors, 0,
            tempered_size - float_array_get(primes, 0));
    }
    else {
        assert(primes->size == generator_mapping->size);
        offset = 0;
    }

    for (int i=0; i<generator_mapping->size; i++) {
        double tempered_size = (double) int_array_get(period_mapping,i+1)
                * this->period_size +
                (double) int_array_get(generator_mapping,i)
                * this->generator_size;
        float_array_set(errors, i+offset,
            tempered_size - float_array_get(primes, i+offset));
    }
    return errors;
}

double linear_temperament_resolve(
        linear_temperament* this,
        int_array* interval) {

    assert(this != NULL);
    assert(interval != NULL);


    int size = interval->size;
    int first_prime = 0;
    double total = 0.0;

    if (size < this->mapping->period_mapping->size) {
        first_prime = 1;
    }
    for (int i=0; i<size; i++) {
        total += (double) this->mapping->period_mapping->data[i+1]
                    * this->period_size +
                    (double) this->mapping->generator_mapping->data[i+1]
                        * this->generator_size;
    }
    return total;
}

void linear_temperament_optimize_rms(linear_temperament *this,
        float_array *primes, generic_list *intervals) {

    assert(this != NULL);
    assert(intervals != NULL);
    assert(primes != NULL);
    double numerator = 0.0;
    double denominator = 0.0;

    double period_tuning = this->period_size;
    int_array* period_mapping = this->mapping->period_mapping;
    int_array* generator_mapping = this->mapping->generator_mapping;
    assert(primes->size == this->mapping->period_mapping->size);

    generic_list *node = intervals;
    while (node != NULL) {
        int_array* interval = (int_array*) generic_list_head(node);
        node = generic_list_tail(node);

        assert(period_mapping->size == interval->size);

        int equivs = 0;
        int gens = 0;
        for (int i=0; i<period_mapping->size; i++) {
            equivs += period_mapping->data[i] * interval->data[i];
            if (i>0) {
                gens += generator_mapping->data[i-1] * interval->data[i];
            }
        }
        if (gens != 0) {
            double optimum = magnitude(primes, interval);
            double gradient = (double) gens;
            double offset = ((double) equivs)*period_tuning - optimum;
            /*
             * y = mx + c
             * y**2 = m**2*x**2 + 2*cmx * c**2
             * d(y**2)/dx = 2*m**2*x + 2cm = 0
             * x = -2cm/2m**2 = -cm/m**2
             */
           numerator -= gradient*offset;
           denominator += gradient*gradient;
       }
    }
   assert (denominator != 0.0);
   this->generator_size = numerator / denominator;
}


double worst_error(float_array *errors, generic_list* intervals) {

    assert(errors != NULL);
    assert(intervals != NULL);

    int_array* interval;
    double error;
    double worst_error = 0.0;
    generic_list* node = intervals;

    while (node != NULL) {
        interval = (int_array*) generic_list_head(node);
        assert(interval != NULL);
        assert(interval->size == errors->size);
        error = fabs(magnitude(errors, interval));
        if (error>worst_error) {
            worst_error = error;
        }
        node = generic_list_tail(node);
    }
    return worst_error;
}

double rms_error(float_array *errors, generic_list *intervals) {

    assert(errors != NULL);
    assert(intervals != NULL);

    double total=0.0;
    generic_list* node = intervals;
    int_array* interval;
    double error;
    int count = 0;

    while (node != NULL) {
        interval = (int_array*) generic_list_head(node);
        assert(interval != NULL);
        assert(interval->size == errors->size);
        error = fabs(magnitude(errors, interval));
        total += error*error;
        count++;
        node = generic_list_tail(node);
    }
    return sqrt(total/count);
}

typedef struct {
    double badness;
    int_array* invariant;
    linear_mapping* mapping;
} lt_element;

lt_element* lt_element_new
        (linear_mapping *mapping, int width, double error) {
    assert(mapping != NULL);
    lt_element* result = malloc(sizeof(lt_element));
    if (result != NULL) {
        long n = int_array_get(mapping->et1, 0);
        long m = int_array_get(mapping->et2, 0);
        if (n>m) {
            long swap = n;
            n = m;
            m = swap;
        }
        result->badness = width*width*error/* + (*/
            /* fudge factor to give a unique ordering for zero error */
            + 1e-10 * width + 1e-15*m + 1.6e-15*n;
        result->invariant = linear_mapping_invariant(mapping);
        assert(result->invariant != NULL);
        result -> mapping = mapping;
    }
    return result;
}

void lt_element_free(lt_element *this) {
    if (this != NULL) {
        if (this->invariant != NULL) {
            int_array_free(this->invariant);
        }
        free(this);
    }
}


/* Insert this new temperament into the list
 * or if it isn't good enough, delete it.
 *
 * A helper function for get_linear_mappings
 */
void insert_or_delete(
        lt_element* lte, lt_element** temperaments,
        int *n_results, int last_result) {

    assert(lte != NULL);
    assert(lte->mapping != NULL);
    assert(temperaments != NULL);

    /* check the badness */
    if (*n_results > last_result) {
        assert(temperaments[last_result] != NULL);
        if (lte->badness > temperaments[last_result]->badness) {
            linear_mapping_free(lte->mapping);
            lt_element_free(lte);
            return; /***/
        }
    }

    /* Do a linear search to see if we already have this temperament.
     *
     * A binary search would be nice, but sometimes the temperament is
     * perfect and the badness is dominated by the sizes of the
     * generating equal temperaments.
     * That means equivalent temperaments can end up in different
     * parts of the list.
     * For full generality, it's best to check the invariants for
     * all of them.
     *
     * The list should be relatively small, so this search isn't so bad.
     */

    /* remember the last temperament with a
     * smaller badness than this one
     * so that we can easily insert it if
     * required
     */
    int last_better = -1;

    for (int i=0; i<*n_results; i++) {
        assert(temperaments[i] != NULL);
        if (int_array_eq(lte->invariant, temperaments[i]->invariant)) {
            /* this temperament is a duplicate
             *
             * because we may as well have the simplest generating ETs
             * (for consistency with other methods where the original ETs
             * are returned in the right order ;)
             * check the badness
             */
             /* see if the full mappings are identical */
            int thesame = int_array_get(lte->mapping->period_mapping, 0)
                == int_array_get(temperaments[i]->mapping->period_mapping, 0);
            int opposite = thesame;
            long per1, per2, gen1, gen2;
            for (int j=1; j<int_array_len(lte->mapping->period_mapping); j++) {
                per1 = int_array_get(lte->mapping->period_mapping, j);
                per2 = int_array_get(temperaments[i]->mapping->period_mapping, j);
                gen1 = int_array_get(lte->mapping->generator_mapping, j-1);
                gen2 = int_array_get(temperaments[i]->mapping->generator_mapping, j-1);
                thesame &= (per1==per2) && (gen1==gen2);
                opposite &= (gen1==-gen2) && (per1+gen1==per2);
            }
            if (!(thesame || opposite)) {
                printf("false match (%li, 0) (%li, 0)",
                    int_array_get(lte->mapping->period_mapping, 0),
                    int_array_get(temperaments[i]->mapping->period_mapping, 0));
                for (int j=1; j<int_array_len(lte->mapping->period_mapping); j++) {
                    printf("[%li, %li] [%li, %li] (%li, %li) (%li, %li) ",
                      int_array_get(lte->mapping->et1, j),
                      int_array_get(lte->mapping->et2, j),
                      int_array_get(temperaments[i]->mapping->et1, j),
                      int_array_get(temperaments[i]->mapping->et2, j),
                      int_array_get(lte->mapping->period_mapping, j),
                      int_array_get(lte->mapping->generator_mapping, j-1),
                      int_array_get(temperaments[i]->mapping->period_mapping, j),
                      int_array_get(temperaments[i]->mapping->generator_mapping, j-1));
                }
                printf("\n");
            }
            if (lte->badness < temperaments[i]->badness) {
                linear_mapping_free(temperaments[i]->mapping);
                lt_element_free(temperaments[i]);
                /* reorder the list */
                while (i>0 && lte->badness < temperaments[i-1]->badness) {
                    temperaments[i] = temperaments[i-1];
                    i--;
                }
                temperaments[i] = lte;
            }
            else {
                linear_mapping_free(lte->mapping);
                lt_element_free(lte);
            }
            return; /***/
        }
        else if (temperaments[i]->badness < lte->badness) {
            last_better = i;
        }
    }
    /* To get to here, the temperament is new and desirable */
    if (*n_results <= last_result) {
        (*n_results)++;
    }
    else {
        assert(last_better < last_result);
        assert(temperaments[last_result] != NULL);
        linear_mapping_free(temperaments[last_result]->mapping);
        lt_element_free(temperaments[last_result]);
    }
    /* move over the temperaments with a higher badness than the new one */
    for (int i=(*n_results)-2; i>last_better; i--) {
        temperaments[i+1] = temperaments[i];
    }
    temperaments[last_better+1] = lte;
}



/* Returns at most the given number of results,
 * although it returns a list rather than an array.
 *
 * If not enough temperaments were found, it'll return an
 * incomplete list.
 *
 * The first result is also specified so that each
 * page in a web app can return a section of the full
 * list.
 */
generic_list* get_linear_mappings(
        generic_list *ets,
        generic_list *consonances,
        float_array *primes,
        long max_complexity,
        double worst_error,
        int first_result, /* indexed from 0 */
        int last_result) {
    assert(primes != NULL);
    assert(max_complexity > 0);
    assert(worst_error >= 0.0);
    assert(first_result >= 0);
    assert(last_result >= first_result);

    lt_element** temperaments = calloc(
            last_result+1,
            sizeof(lt_element *));
    assert(temperaments != NULL);
    int n_results = 0;
    generic_list* consist = ets;
    while (consist != NULL) {
        equal_temperament* et1 = equal_temperament_list_head(consist);
        assert(et1 != NULL);
        consist = generic_list_tail(consist);
        generic_list *remainder = consist;
        while(remainder != NULL) {
            equal_temperament* et2 = equal_temperament_list_head(remainder);
            assert(et2 != NULL);
            remainder = generic_list_tail(remainder);
            linear_mapping *mapping = linear_mapping_new(
                    et1->mapping, et2->mapping);
            long complexity =
                linear_mapping_max_width(mapping, consonances)
                * int_array_get(mapping->period_mapping, 0);
            long hcf = int_array_get(mapping->generator_mapping, 0);
            for (int i=1; i<mapping->generator_mapping->size; i++) {
                hcf = gcd(hcf, int_array_get(mapping->generator_mapping, i));
            }
            if (complexity > max_complexity || hcf > 1) {
                linear_mapping_free(mapping);
            }
            else {
                linear_temperament* lt = linear_temperament_new(mapping, 0.0);
                linear_temperament_optimize_rms(lt, primes, consonances);
                float_array* errors = linear_temperament_errors(lt, primes);
                linear_temperament_free(lt);
                double error = rms_error(errors, consonances);
                float_array_free(errors);
                if (error >= worst_error) {
                    linear_mapping_free(mapping);
                }
                else {
                    lt_element* lte =
                        lt_element_new(mapping, complexity, error);
                    assert(lte != NULL);
                    insert_or_delete(
                            lte, temperaments, &n_results, last_result);
                }
            }
        }
    }

    /* now we have an array, but we need to return a list */
    generic_list* result = generic_list_new();
    double last_badness = 1e100; /* just a check */
    for (int i=first_result; i<n_results; i++) {
        /* build the list in reverse order */
        lt_element* lte = temperaments[n_results-1-i];
        assert(lte != NULL);
        result = generic_list_add(result, lte->mapping);
        assert(lte->badness <= last_badness);
        last_badness = lte->badness;
    }
    for (int i=0; i<n_results; i++) {
        if(i<first_result) {
            linear_mapping_free(temperaments[i]->mapping);
        }
        lt_element_free(temperaments[i]);
    }
    free(temperaments);
    return result;
}


/*
 *
 * loads of array types that Python can access
 *
 *
 * type names follow Python rather than C
 *
 */

int_array* int_array_new(int size) {
    assert(size>0);
    int_array* this = (int_array*) malloc(sizeof(int_array));
    if (this != NULL) {
        this->data = (long*) calloc(size, sizeof(long));
        if (this->data == NULL) {
            this->size = 0;
        }
        else {
            this->size = size;
        }
    }
    return this;
}

void int_array_set(int_array* this, int index, long value) {
    assert(this != NULL);
    assert(0 <= index);
    assert(index < this->size);
    this->data[index] = value;
}

long int_array_get(int_array* this, int index) {
    assert(this != NULL);
    assert(0 <= index);
    assert(index < this->size);
    return this->data[index];
}

int int_array_len(int_array* this) {
    assert(this != NULL);
    return this->size;
}

/* say if the two arrays are equal */
int int_array_eq(int_array* this, int_array *other) {
    int len = int_array_len(this);
    if (len != int_array_len(other)) {
        return 0; /***/
    }
    for (int i=0; i<len; i++) {
        if (int_array_get(this, i) != int_array_get(other, i)) {
            return 0; /***/
        }
    }
    return 1;
}


void int_array_free(int_array* this) {
    if(this != NULL) {
        if (this->data != NULL) {
            free(this->data);
        }
        free(this);
    }
}


float_array* float_array_new(int size) {
    assert(size>0);
    float_array* this = (float_array*) malloc(sizeof(float_array));
    if (this != NULL) {
        this->data = (double*) calloc(size, sizeof(double));
        if (this->data == NULL) {
            this->size = 0;
        }
        else {
            this->size = size;
        }
    }
    return this;
}

void float_array_set(float_array* this, int index, double value) {
    assert(this != NULL);
    assert(0 <= index);
    assert(index < this->size);
    this->data[index] = value;
}

double float_array_get(float_array* this, int index) {
    assert(this != NULL);
    assert(0 <= index);
    assert(index < this->size);
    return this->data[index];
}

int float_array_len(float_array* this) {
    assert(this != NULL);
    return this->size;
}

void float_array_free(float_array* this) {
    if (this != NULL) {
        if (this->data != NULL) {
            free(this->data);
        }
        free(this);
    }
}

generic_list* generic_list_new(void) {
    return (generic_list*) NULL;
}

generic_list* generic_list_add(generic_list* old, void* item) {
    generic_list* this = (generic_list*) malloc(sizeof(generic_list));
    assert(this != NULL);
    this->item = item;
    this->next = old;
    return this;
}

void* generic_list_head(generic_list* this) {
    assert(this != NULL);
    return this->item;
}

generic_list* generic_list_tail(generic_list* this) {
    /* doesn't free the original element */
    assert(this != NULL);
    return this->next;
}

void generic_list_free(generic_list* this) {
    if (this != NULL) {
        generic_list_free(this->next);
        free(this);
    }
}



