(********************************************************************

To load this into the interpreter:

#directory "d:/develop";;
#load "temper.cmo";;

********************************************************************)


let log2 x = log(x)/.log(2.)
let prime_numbers = [2;3;5;7;11;13;17;19]
let standard_primes = List.map log2 (
    List.map float_of_int (List.tl prime_numbers))

exception Temperament_exception

let magnitude primes interval =
    List.fold_left ( +.) 0.0
        (List.map2 ( *.) (List.map float_of_int interval) primes)

let nint x = int_of_float (floor (x +. 0.5))

let rec factorize prime_nums n =
    (* returns a list the same size as prime_nums *)
    match prime_nums, n with
          [], 1 -> []
        | [], _ -> raise Not_found
        | prime::others, 1 ->
            0::(factorize others 1) (* could be more efficient *)
        | prime::others, n ->
            if n<1 then
                invalid_arg "positive numbers only"
            else
                let rec count_factors residue =
                    if residue mod prime = 0 then
                        let count, sub_residue =
                            count_factors (residue/prime) in
                        succ count, sub_residue
                else
                    0, residue
                in
                let factors, residue = (count_factors n) in
                factors :: (factorize others residue)

let factorize_number = factorize prime_numbers

let factorize_ratio = function n, d ->
    List.map2 (-)
        (factorize_number n)
        (factorize_number d)

let rec conditional_combinations number input condition =
  (*
  Return all combinations of number elements from the input list.
  The condition is a function returning a boolean to decide
  whether a partial result is valid.
  This can be useful when a large proportion of the results
  are useless, and this can be predicted early on.
  *)
    if number=1 then
        List.filter condition
            (List.map (function element -> [element]) input)
    else if number>0 then
        let rec inner_function counter input2 =
            if counter < 0 then
                []
            else
                let remainder = List.tl input2 in
                List.append
                    (List.filter condition
                        (List.map
                            (List.append [List.hd input2])
                            (conditional_combinations
                                (number-1) remainder condition)))
                    (inner_function (counter - 1) remainder)
        in
        inner_function (List.length input - number) input
    else
        invalid_arg "can't return a set with negative size"

let combinations number input =
    conditional_combinations number input (function _ -> true)


(********************************************************************
*
*  Tonality diamonds
*
********************************************************************)


class interval_set intervals_in primes_in =
    (* this is octave equivalent,
     * along with intervals whenever they're
     * used for errors or optimizations
     *)
    let rec truncate new_size old_list =
        match new_size, old_list with
              0, _ -> []
            | n, x::xs when n>0 -> x::(truncate (pred n) xs)
            | _ -> invalid_arg "new_size must be at least as big as the list is already"
    in
    let simplify_intervals intervals =
        let dimension intervals =
            let rec last_non_zero = function
                  [] -> 0
                | x::xs ->
                    if last_non_zero xs > 0 then
                        last_non_zero xs + 1
                    else if x=0 then 0
                    else 1
            in
            List.fold_left max 0 (List.map last_non_zero intervals);
        in
        List.map (truncate (dimension intervals)) intervals
    in
    let intervals = simplify_intervals intervals_in in
    let primes = truncate
            (List.length (List.hd intervals)) primes_in
    in
    object (self)
        method intervals = intervals
        method primes = primes
        method magnitudes = List.map (magnitude primes) intervals

        method cross_set =
            let rec cross_set_rec = function [] -> []
                | [harm1;harm2]::remainder ->
                    let new_interval = List.map2 (-) harm1 harm2
                    and story_so_far = cross_set_rec remainder
                    in
                    if List.mem new_interval story_so_far then
                        story_so_far
                    else
                        new_interval::story_so_far
                | _ -> failwith "really, this should never happen"
            in
            new interval_set
                (cross_set_rec (combinations 2 intervals))
                primes
            end

let odd_harmonics n =
    let rec odd_harmonics_rec = function
        1 -> [List.tl (factorize_number 1)]
        | n when (n mod 2)=1 ->
            (List.tl (factorize_number n))::(odd_harmonics_rec (n-2))
        | _ -> invalid_arg "positive odd integers only"
    in
    new interval_set (odd_harmonics_rec n) standard_primes

let odd_limit n = (odd_harmonics n)#cross_set

let harmonic_diamond harmonics size =
    let rec identity_array n = function
          0 -> []
        | m ->
            let rec each_row row = function
                  0 -> []
                | n -> (if n=row then 1 else 0) :: each_row row (n-1)
            in
            each_row m n :: identity_array n (m-1)
    in
    (new interval_set (identity_array size (size+1)) harmonics)#cross_set


(********************************************************************
*
*  Equal Temperaments
*
********************************************************************)

class equal_temperament mapping =
  (* mapping contains the octave *)
  object (self)

    method mapping = mapping

    method errors primes =
        (* the difference between each prime and
           its ideal value, where octave=1.0 now
         *)
        let octave = float_of_int (List.hd mapping) in
        let rec errors_for mapping primes =
            match mapping, primes with
                [], [] -> []
              | map::other_maps, prime::other_primes ->
                    (float_of_int map /. octave -. prime)
                    :: errors_for other_maps other_primes
              | _ -> invalid_arg "lists are different lengths"
        in
            errors_for (List.tl mapping) primes

    method resolve = magnitude
        (1.0::(List.map
            (function x -> (float_of_int x) /.
                 (float_of_int (List.hd mapping)))
            (List.tl mapping)))

    method worst_error (intervals:interval_set) =
        List.fold_left max 0.0 (List.map abs_float
            (List.map
                (magnitude (self#errors intervals#primes))
                intervals#intervals))

    method rms_error (intervals:interval_set) =
        let get_error = magnitude (self#errors intervals#primes) in
        sqrt (
            (List.fold_left (+.) 0.0
                (List.map
                    (function interval ->
                        (get_error interval)**2.)
                    intervals#intervals))
            /. float_of_int (List.length intervals#intervals))
  end

let alternative_mappings et primes cutoff =
  (* Get all equal temperaments that might be consistent
   * I don't think cutoffs higher than 1.0 are meaningful
   *)
    let rec alternative_mappings_rec et index =
        if index = List.length et#mapping then
            [et]
        else
            let result = alternative_mappings_rec
                et (index+1)
            and target = List.nth primes (index-1) *.
                float_of_int (List.hd et#mapping)
            and approximation = float_of_int
                (List.nth et#mapping index)
            in
            let error = approximation -. target
            in
            if (1.0 -. abs_float error) < cutoff then
                let basis = Array.of_list et#mapping
                in (
                    basis.(index) <- basis.(index) +
                        if error < 0.0 then 1 else -1;
                    result @ alternative_mappings_rec
                        (new equal_temperament
                            (Array.to_list basis))
                            (index+1)
                    )
            else result
    in
    alternative_mappings_rec et 1


let prime_et n_notes primes =
  (* return the equal temperament with
     the right number of notes to the
     (formal) octave
     and the nearest approximation to
     each prime interval
  *)
    if n_notes>0 then
        new equal_temperament (
            n_notes::List.map nint
                (List.map (( *.) (float_of_int n_notes)) primes))
    else
        invalid_arg "number of notes must be positive"

let best_et n_notes consonances =
    snd (List.hd (List.sort compare (List.map
        (function et ->
            (et#worst_error consonances, et))
        (alternative_mappings
            (prime_et n_notes consonances#primes)
            consonances#primes 1.0)
        )))


let rec gcd m = function
      0 -> abs m
    | n -> gcd n (m mod n)


let get_limited_ets consonances n_ets cutoff =
  (*
  Get n_ets EqualTemperaments consistent in the given consonance limit

  The cutoff is the largest allowable error in terms of step sizes,
  for a consistent mapping.
  0.5 is the value for strict consistency.
  May return some bonus ETs!
  *)
    let rec try_next n_notes old_candidates =
        let pet = prime_et n_notes consonances#primes in
        let candidates = old_candidates @ List.filter
            (function et ->
                List.fold_left gcd 0 et#mapping = 1
                &&
                et#worst_error consonances
                    *. float_of_int (List.hd et#mapping)
                        < cutoff)
            (if cutoff<=0.5 then
                [pet]
            else
                alternative_mappings (pet) consonances#primes cutoff)
        in
        if List.length candidates < n_ets then
            try_next (n_notes+1) candidates
        else candidates
    in
    try_next 1 []

(********************************************************************
*
*  Linear Temperaments
*
********************************************************************)

class linear_mapping primes_by_melody =
    let m, n = (List.hd primes_by_melody) in
    let m_gen, n_gen, octave_division =
        (* Extended Euclidian algorithm
           from von zur Gathen & Gerhard p.47
         *)
        let rec extended_euclid
                (last_t, last_s, last_r)
                (next_t, next_s, next_r) =
            if next_r=0 then
                last_t, last_s, last_r
            else
                let q = last_r / next_r in
                let r = last_r mod next_r in
                let sign_of_r = if r<0 then -1 else 1 in
                extended_euclid (next_t, next_s, next_r)
                    ((last_t - q*next_t) / sign_of_r,
                        (last_s - q*next_s) / sign_of_r,
                        (abs r))
        in
        if m<0 or n<0 then
            invalid_arg "only positive numbers of steps per octave"
        else
            extended_euclid (0, 1, m) (-1, 0, n)
    in
    let mapping =
        (* get primes in terms of equivalence interval and generator *)
        assert (abs(m*n_gen - n*m_gen) = octave_division);
        assert (octave_division>0);
        (octave_division, 0)::(List.map
            (function axis1, axis2 ->
                (n_gen*axis1 - m_gen*axis2),
                (m*axis2 - n*axis1) / octave_division
                * (if m_gen<0 then -1 else 1))
            (List.tl primes_by_melody))
        (* if the intervals aren't orthogonal,
           List.filter ((<>) 0) (snd (List.split result)) = []
         *)
    in
    let rec resolve_mapping = function
      (* labels are for equivalences and generators,
         but it's more general than that
       *)
          [], [] -> 0, 0
        | prime::mapped, (equiv_map, gen_map)::mapping ->
            let equivs_to_interval, gens_to_interval =
                resolve_mapping(mapped, mapping)
            in
            (prime*equiv_map + equivs_to_interval),
            (prime*gen_map + gens_to_interval)
        | _ -> invalid_arg "the interval doesn't have the expected size (it should be octave specific)"
    in
  object (self)

    method primes_by_melody = primes_by_melody
    method generator = abs(n_gen + m_gen)
    method mapping = mapping
    method periods_per_octave = octave_division

    method generator_mapping =
        snd (List.split (List.tl self#mapping))

    method invariant =
      (* something that doesn't depend on primes_by_melody
         or the generator chosen
       *)
        self#periods_per_octave,
        let gen_map = self#generator_mapping in
        if List.hd (List.filter ((<>) 0) gen_map) < 0 then
            List.map (~-) gen_map
        else
            gen_map

    method width interval =
        abs (List.fold_left (+) 0
            (List.map2 ( * ) interval self#generator_mapping))

    method widest = function
          [] -> invalid_arg "come on, you have to supply at least one interval"
        | [interval] -> interval
        | interval::others ->
            let other = self#widest others in
            if self#width interval > self#width other then
                interval
            else
                other

    method max_width intervals = self#width (self#widest intervals)

    method by_melody mapped =
        resolve_mapping(mapped, primes_by_melody)

    method by_mapping mapped =
        resolve_mapping(mapped, self#mapping)

    method rms_complexity intervals =
        let ppo = float_of_int self#periods_per_octave
        in
        sqrt (
            List.fold_left (+.) 0.0
                (List.map
                    (function interval ->
                        (float_of_int (self#width interval)
                            *. ppo)**2.)
                    intervals)
            /. float_of_int (List.length intervals - 1))
  end

let linear_from_equals et1 et2 =
    new linear_mapping (List.combine et1#mapping et2#mapping)

class linear_temperament melodic_mapping generator_size =
  (*
   * a temperament is a mapping with a specific tuning
   * in this case, it has to have perfect octaves
   * (whatever they are)
   * because that's how the optimizations work
   *
   * you'd better get the right generator as well
   * (always the smallest of all positive equivalents,
   *  according to the ET mappings you supply)
   *)
  object (self)
    inherit linear_mapping melodic_mapping
    val mutable generator_tuning = generator_size

    method tuning = 1.0 /. float_of_int self#periods_per_octave,
        generator_tuning

    method resolve interval =
        let by_per, by_gen = (self#by_mapping interval)
        and per_size, gen_size = self#tuning
        in
        per_size *. float_of_int by_per +.
            gen_size *. float_of_int by_gen

    method errors primes =
        (* the difference between each prime and
           its ideal value, in octaves
         *)
        let rec errors_for mapping primes =
            let per_size, gen_size = self#tuning
            in
            match mapping, primes with
                [], [] -> []
              | (map1, map2)::other_maps, prime::other_primes ->
                    ((float_of_int map1 *. per_size +.
                            float_of_int map2 *. gen_size)
                        -. prime)
                    :: errors_for other_maps other_primes
              | _ -> invalid_arg "lists are different lengths"
        in
            errors_for (List.tl self#mapping) primes

    method smallest_containing_mos intervals =
        (* Dave Keenan's algorithm, propriety not enforced *)
        let width = self#max_width intervals + 1
        and ratio = ref (abs_float generator_tuning *.
            float_of_int self#periods_per_octave)
        in
        let i = ref (floor !ratio)
        and m_prev, m = ref 0, ref 1
        and swap x y =
            let temp = !x in
            x := !y;
            y := temp;
            ()
        in
        while !m < width do
            if !ratio  = !i then failwith "division by zero ahead!";
            ratio := 1.0 /. (!ratio -. !i);
            i := floor !ratio;
            if !i = 0. then failwith "i zero";
            swap m m_prev;
            m := !m + !m_prev * (int_of_float !i);
            assert (!m_prev > 0);
            if !m > width then
                (* check we didn't overshoot *)
                m := width + ((!m-width) mod !m_prev)
        done;
        !m * self#periods_per_octave

    method worst_error (intervals:interval_set) =
        (* copied from equal_temperament *)
        List.fold_left max 0.0 (List.map abs_float
            (List.map
                (magnitude (self#errors intervals#primes))
                intervals#intervals))

    method rms_error (intervals:interval_set) =
        (* another one copied from equal_temperament *)
        let get_error = magnitude (self#errors intervals#primes) in
        sqrt (
            (List.fold_left (+.) 0.0
                (List.map
                    (function interval ->
                        (get_error interval)**2.)
                    intervals#intervals))
            /. float_of_int (List.length intervals#intervals))

    method optimize_minimax (consonances:interval_set) =
        let depends_on_generator interval =
                snd (self#by_mapping (0::interval)) <> 0
        in
        let optimizing_set = List.filter
            depends_on_generator consonances#intervals
        in
        if optimizing_set = [] then
            raise Temperament_exception
        else
            let make_interval_exact =
                self#make_interval_exact consonances#primes
            and get_max_error() =
                self#worst_error (new interval_set
                    optimizing_set
                    consonances#primes)
            in
            let rec get_best_and_minimax = function
                  [] -> [], 1e150
                | interval::others ->
                    let best_intervals, minimax =
                        get_best_and_minimax others
                    in
                    make_interval_exact interval;
                    let max_error = get_max_error() in
                    if abs_float(max_error -. minimax) < 1e-12 then
                        (* near enough equal *)
                        interval::best_intervals, minimax
                    else if max_error < minimax then
                        [interval], max_error
                    else
                        best_intervals, minimax
            in
            let best, minimax = get_best_and_minimax optimizing_set in
            match best with
                  [] -> failwith "minimax overflow"
                | [interval] -> make_interval_exact interval
                | intervals ->
                    (* even this will only work for
                       two equally bad optimizations
                       but it's a fairly good bet
                     *)
                    let best_pairs = combinations 2 intervals
                    and combine_with operation = function
                        [a; b] -> List.map2 operation a b
                        | _ -> failwith "not pairs!"
                    in
                    let best, minimax = get_best_and_minimax (
                        List.filter depends_on_generator (
                            List.hd best ::
                            List.map (combine_with (+)) best_pairs @
                            List.map (combine_with (-)) best_pairs))
                    in
                    make_interval_exact (List.hd best)


    method optimize_rms (consonances:interval_set) =
        let rec get_fraction = function
              [] -> 0.0, 0.0
            | interval::others ->
                let equivs, gens
                    = self#by_mapping (0::interval)
                in
                if gens=0 then get_fraction others
                  (* there's no point in optimizing over
                     intervals that don't depend on the
                     mapping
                   *)
                else
                    let optimum = magnitude consonances#primes interval
                    and gradient = float_of_int gens
                    and period_tuning = fst self#tuning in
                    let offset = float_of_int
                        equivs*.period_tuning -. optimum
                        (*
                        y = mx + c
                        y**2 = m**2*x**2 + 2*cmx + c^^2
                        d(y**2)/dx = 2*m**2*x + 2cm = 0
                        x = -2cm/2m**2 = -cm/m**2
                        *)
                    and numerator, denominator = get_fraction others
                    in
                    numerator -. gradient *. offset,
                    denominator +. gradient**2.
        in
        let numerator, denominator = get_fraction consonances#intervals in
        if denominator = 0.0 then
            raise Temperament_exception
        else
            generator_tuning <- numerator /. denominator


    method make_interval_exact primes interval =
        let equivs_to_interval, gens_to_interval =
            self#by_mapping (0::interval)
        in
        if gens_to_interval = 0 then
            raise Temperament_exception
        else
            generator_tuning <- (
                    magnitude primes interval -.
                    float_of_int equivs_to_interval /.
                    float_of_int self#periods_per_octave)
                 /. float_of_int gens_to_interval

  end

let linearize_equals et1 et2 =
    let map_obj = linear_from_equals et1 et2 in
    new linear_temperament
        map_obj#primes_by_melody
        (float_of_int map_obj#generator /.
            float_of_int (List.hd et1#mapping + List.hd et2#mapping))

let best_linear m n consonances =
    let et = linearize_equals
        (best_et m consonances)
        (best_et n consonances)
    in
        et#optimize_rms consonances;
        et

let get_linear_temperaments ets intervals
        figure_of_demerit highest_complexity worst_error use_rms =

    let primes = intervals#primes
    and candidates = Hashtbl.create 100
    in
    begin
        List.iter (function
              [et1; et2] ->
                let linmap = linear_from_equals et2 et1 in
                let complexity =
                    linmap#max_width intervals#intervals *
                        linmap#periods_per_octave
                in
                if complexity <= highest_complexity &&
                    List.fold_left gcd 0 linmap#generator_mapping = 1
                then
                    try
                        let tuning = new linear_temperament
                            linmap#primes_by_melody 0.
                        in
                        let error =
                            if use_rms then
                            begin
                                tuning#optimize_rms intervals;
                                tuning#rms_error intervals
                            end
                            else
                            begin
                                tuning#optimize_minimax intervals;
                                tuning#worst_error intervals
                            end
                        in
                        if error <= worst_error then
                            let key = tuning#invariant in
                            if not (Hashtbl.mem candidates key) then
                                Hashtbl.add candidates key
                                    (figure_of_demerit tuning, tuning)
                    with Temperament_exception ->
                        ()
                else ()
            | _ -> failwith "Pah!  It should always have a 2 element list"
            ) (combinations 2 ets);
        let decorated =
            Hashtbl.fold
                (fun _ item others ->item::others)
            candidates []
        in
        let sorted = List.sort
            (fun (x, _) (y, _) -> compare x y)
            decorated
        in
        List.map snd sorted
    end

let standard_figure_of_demerit intervals lt =
    let error = lt#rms_error intervals
    and width = lt#max_width intervals#intervals
         * lt#periods_per_octave
    in
         float_of_int (width*width) *. error
         (* add a fudge factor to give a unique ordering *)
         +. 1e-10 *.
             (let m, n = List.hd lt#primes_by_melody in
             1.618 *. float_of_int m +. float_of_int n)

let minimax_figure_of_demerit intervals lt =
    let error = lt#worst_error intervals
    and width = lt#max_width intervals#intervals
         * lt#periods_per_octave
    in
         float_of_int (width*width) *. error
         (* add a fudge factor to give a unique ordering *)
         +. 1e-10 *.
             (let m, n = List.hd lt#primes_by_melody in
             1.618 *. float_of_int m +. float_of_int n)

