"""Wrapper around the wrapper around the C temperament library.

Could probably wrap _temper_c directly as it stands.
"""

import temper_c

class abstract_array:
    def __init__(self, seq, garbage_collect=1):
        try:
            length = len(seq)
            data = self.array_new(length)
            for i in range(length):
                self.array_set(data, i, seq[i])
        except TypeError:
            data = seq
        self.data = data
        self.garbage_collect = garbage_collect

    def __len__(self):
        return self.array_len(self.data)

    def __getitem__(self, index):
        if index<0 or index>=len(self):
            raise IndexError("list index out of range")
        return self.array_get(self.data, index)

    def __setitem__(self, index, value):
        if index<0 or index>=len(self):
            raise IndexError("list index out of range")
        self.array_set(self.data, index, value)

    def __del__(self):
        if self.garbage_collect:
            self.array_free(self.data)

class int_array(abstract_array):
    array_new = temper_c.int_array_new
    array_set = temper_c.int_array_set
    array_get = temper_c.int_array_get
    array_len = temper_c.int_array_len
    array_free = temper_c.int_array_free

class float_array(abstract_array):
    array_new = temper_c.float_array_new
    array_set = temper_c.float_array_set
    array_get = temper_c.float_array_get
    array_len = temper_c.float_array_len
    array_free = temper_c.float_array_free

class generic_list:
    def __init__(self, seq=()):
        data = temper_c.generic_list_new()
        for item in seq:
            data = temper_c.generic_list_add(data, item)
        self.top = self.current = data

    def add(self, item):
        new_top = temper_c.generic_list_add(self.top, item)
        assert new_top, "Couldn't allocate new item"
        self.top = self.current = new_top

    def next(self):
        if self.current is None:
            raise IndexError, "no more"
        result = temper_c.generic_list_head(self.current)
        self.current = temper_c.generic_list_tail(self.current)
        return result

    def __del__(self):
        temper_c.generic_list_free(self.top)

class equal_temperament:
    temper_c = temper_c # the global module gets collected too early

    def __init__(self, pointer):
        self.data = pointer
        self.mapping = int_array(
            temper_c.equal_temperament_mapping(pointer))

    def errors(self, primes):
        primes_array = float_array(primes)
        return tuple(float_array(
            temper_c.equal_temperament_errors(
                self.data, primes_array.data)))

    def resolve(self, interval):
        interval_array = int_array(interval)
        return temper_c.equal_temperament_resolve(
            self.data, interval_array.data)

    def worst_error(self, intervals, primes):
        primes_array = float_array(primes)
        interval_list = generic_list()
        interval_cache = [] # something to hold reference counts
        for interval in intervals:
            interval_array = int_array(interval)
            interval_cache.append(interval_array)
            interval_list.add(interval_array.data)
        return temper_c.equal_temperament_worst_error(
            self.data, interval_list.top, primes_array.data)

    def __del__(self):
        # if this gets called when the script exits,
        # temper_c has already been garbage collected
        # so we need to use our own copy
        self.temper_c.equal_temperament_free(self.data)


def magnitude(primes, interval):
    primes_array = float_array(primes)
    interval_array = int_array(interval)
    return temper_c.magnitude(
        primes_array.data,
        interval_array.data)

def alternative_mappings(et, primes, cutoff):
    prime_array = float_array(primes)
    et_list = temper_c.alternative_mappings(et.data, prime_array.data, cutoff)
    result = []
    node = et_list
    while node is not None:
        result.append(equal_temperament(
            temper_c.equal_temperament_list_head(node)))
        node = temper_c.generic_list_tail(node)
    temper_c.generic_list_free(et_list)
    return result

def prime_et(n_notes, primes):
    prime_array = float_array(primes)
    return equal_temperament(temper_c.prime_et(n_notes, prime_array.data))

def best_et(n_notes, consonances, primes):
    consonance_list = generic_list()
    consonance_cache = [] # something to hold reference counts
    for interval in consonances:
        interval_array = int_array(interval)
        consonance_cache.append(interval_array)
        consonance_list.add(interval_array.data)
    prime_array = float_array(primes)

    return equal_temperament(
        temper_c.best_et(
            n_notes, consonance_list.top, prime_array.data))

def get_limited_ets(consonances, primes, n_ets, cutoff):
    consonance_list = generic_list()
    consonance_cache = [] # something to hold reference counts
    for interval in consonances:
        interval_array = int_array(interval)
        consonance_cache.append(interval_array)
        consonance_list.add(interval_array.data)
    prime_array = float_array(primes)

    et_list = temper_c.get_limited_ets(
            consonance_list.top, prime_array.data, n_ets, cutoff)

    result = []
    node = et_list
    while node is not None:
        result.append(equal_temperament(
            temper_c.equal_temperament_list_head(node)))
        node = temper_c.generic_list_tail(node)
    temper_c.generic_list_free(et_list)
    return result



class linear_mapping:
    temper_c = temper_c # the global module gets collected too early

    def __init__(self, mapping):
        self.data = mapping
        self.deep_copy = 1

    def et1(self):
        return tuple( int_array(
            temper_c.linear_mapping_et1(self.data), 0))

    def et2(self):
        return tuple( int_array(
            temper_c.linear_mapping_et2(self.data), 0))

    def period_mapping(self):
        return tuple( int_array(
            temper_c.linear_mapping_period_mapping(self.data), 0))

    def generator_mapping(self):
        return tuple( int_array(
            temper_c.linear_mapping_generator_mapping(self.data), 0))

    def invariant(self):
        return tuple( int_array(
            temper_c.linear_mapping_invariant(self.data), 1))


    def width(self, interval):
        interval_array = int_array(interval)
        return temper_c.linear_mapping_width(self.data, interval_array.data)

    def widest(self, intervals):
        interval_list = generic_list()
        interval_cache = [] # something to hold reference counts
        for interval in intervals:
            interval_array = int_array(interval)
            interval_cache.append(interval_array)
            interval_list.add(interval_array.data)
        return tuple(int_array(temper_c.linear_mapping_widest(
            self.data, interval_list.top), 0))

    def max_width(self, intervals):
        interval_list = generic_list()
        interval_cache = [] # something to hold reference counts
        for interval in intervals:
            interval_array = int_array(interval)
            interval_cache.append(interval_array)
            interval_list.add(interval_array.data)
        return temper_c.linear_mapping_max_width(
            self.data, interval_list.top)

    def __del__(self):
        # if this gets called when the script exits,
        # temper_c has already been garbage collected
        # so we need to use our own copy
        if self.deep_copy:
            self.temper_c.linear_mapping_free(self.data)


class linear_temperament(linear_mapping):
    def __init__(self, temperament, generator=0.5):
        if isinstance(temperament, linear_mapping):
            self.temperament_data = temper_c.linear_temperament_new(
                    temperament.data, generator)
        else:
            self.temperament_data = temperament
            linear_mapping.__init__(self,
                temper_c.linear_temperament_as_linear_mapping(temperament))

    def as_linear_mapping(self):
        mapping = linear_mapping(
                temper_c.linear_temperament_as_linear_mapping
                    (self.temperament_data))
        mapping.deep_copy = 0 # prevent erroneous deletion
        return mapping

    def errors(self, primes):
        primes_array = float_array(primes)
        return tuple(float_array(
            temper_c.linear_temperament_errors(
                self.temperament_data, primes_array.data)))

    def resolve(self, interval):
        interval_array = int_array(interval)
        return temper_c.linear_temperament_resolve(
            self.temperament_data, interval_array.data)

    def worst_error(self, primes, intervals):
        primes_array = float_array(primes)
        interval_list = generic_list()
        interval_cache = [] # something to hold reference counts
        for interval in intervals:
            interval_array = int_array(interval)
            interval_cache.append(interval_array)
            interval_list.add(interval_array.data)
        errors = float_array(temper_c.linear_temperament_errors(
                self.temperament_data, primes_array.data))
        return temper_c.worst_error(errors.data, interval_list.top)

    def optimize_rms(self, primes, intervals):
        primes_array = float_array([1.0] + list(primes))
        interval_list = generic_list()
        interval_cache = [] # something to hold reference counts
        for interval in intervals:
            interval_array = int_array(temper.reduceWithinTritone(interval))
            interval_cache.append(interval_array)
            interval_list.add(interval_array.data)
        temper_c.linear_temperament_optimize_rms(self.temperament_data,
                primes_array.data, interval_list.top)

    def rms_error(self, primes, intervals):
        primes_array = float_array(primes)
        interval_list = generic_list()
        interval_cache = [] # something to hold reference counts
        for interval in intervals:
            interval_array = int_array(interval)
            interval_cache.append(interval_array)
            interval_list.add(interval_array.data)
        errors = float_array(temper_c.linear_temperament_errors(
                self.temperament_data, primes_array.data))
        return temper_c.rms_error(errors.data, interval_list.top)

    def __del__(self):
        # if this gets called when the script exits,
        # temper_c has already been garbage collected
        # so we need to use our own copy
        self.temper_c.linear_temperament_free(self.temperament_data)

def get_linear_mappings(ets, consonances, primes,
        highest_width, worst_error, first_result, last_result):

    et_list = generic_list()
    et_cache = [] # something to hold reference counts
    for et in ets:
        et_cache.append(et)
        et_list.add(et.data)

    consonance_list = generic_list()
    consonance_cache = [] # something to hold reference counts
    for interval in consonances:
        interval_array = int_array(interval)
        consonance_cache.append(interval_array)
        consonance_list.add(interval_array.data)
    prime_array = float_array([1.0] + list(primes))

    lt_list = temper_c.get_linear_mappings(
            et_list.top, consonance_list.top, prime_array.data,
            highest_width, worst_error, first_result, last_result)

    result = []
    node = lt_list
    while node is not None:
        result.append(linear_mapping(
            temper_c.linear_mapping_list_head(node)))
        node = temper_c.generic_list_tail(node)
    temper_c.generic_list_free(lt_list)
    return result


def test_examples(temper, primes):
    float_tolerance = 1e-10
    h12 = temper.PrimeET(12, primes)
    ch12 = prime_et(12, primes)
    assert tuple(h12.basis) == tuple(ch12.mapping)

    h21 = temper.BestET(21, temper.limit11)
    ch21 = best_et(21, temper.limit11, temper.limit11.primes)
    assert tuple(h21.basis) == tuple(ch21.mapping)

    for err1, err2 in map(None, h12.getErrors(), ch12.errors(primes)):
        assert abs(err1-err2)<float_tolerance

    min3rd = (1, 1, -1)
    assert int(ch12.resolve(min3rd)*h12.basis[0]) == temper.dotprod(min3rd, h12.basis)
    assert magnitude(primes, min3rd) == temper.dotprod(min3rd, [1.0]+primes)

    assert abs(
        h12.getWorstError(temper.limit5) - 
        ch12.worst_error(temper.limit5, primes)) < float_tolerance

    calternatives = {}
    for et in alternative_mappings(ch12, primes, 1.0):
        mapping = tuple(et.mapping)
        assert not calternatives.has_key(mapping)
        calternatives[mapping] = 1

    alternatives = {}
    for et in temper.alternativeMappings(h12, 1.0):
        mapping = tuple(et.basis)
        assert not alternatives.has_key(mapping)
        alternatives[mapping] = 1

    assert calternatives==alternatives

def test_linear_examples(temper, primes, m, n, consonances):
    et1 = temper.BestET(m, consonances)
    et2 = temper.BestET(n, consonances)
    basis1 = int_array(et1.basis) # save from garbage collection
    basis2 = int_array(et2.basis)

    cet1 = best_et(m, consonances, consonances.primes)
    cet2 = best_et(n, consonances, consonances.primes)
    clinmap = linear_mapping(
            temper_c.linear_mapping_new(cet1.mapping.data,
                cet2.mapping.data))
    assert tuple(clinmap.et1()) == tuple(basis1)
    assert tuple(clinmap.et2()) == tuple(basis2)

    lintemp = et1&et2
    assert lintemp.mapping == map(None,
        clinmap.period_mapping(),
        (0, )+clinmap.generator_mapping())
    invariant = lintemp.invariant()
    if filter(None, invariant[1:])[0] < 0:
        import operator
        invariant = invariant[:1] + tuple(map(operator.neg, invariant[1:]))
    assert clinmap.invariant() == invariant

    intervals = []
    for interval in consonances:
        intervals.append(temper.reduceWithinTritone(interval))
        assert lintemp.getWidth(interval) == clinmap.width(intervals[-1])
    assert lintemp.getWidth(lintemp.getWidestInterval()
        ) == clinmap.max_width(intervals)
    assert clinmap.max_width(intervals) == clinmap.width(
        clinmap.widest(intervals))

    lintemp.optimizeRMS()
    clintemp = linear_temperament(clinmap)

    clinmap_dup = clintemp.as_linear_mapping()
    assert clinmap_dup.invariant() == invariant
    assert tuple(clinmap_dup.et1()) == tuple(basis1)
    assert tuple(clinmap_dup.et2()) == tuple(basis2)
    assert lintemp.mapping == map(None,
        clinmap_dup.period_mapping(),
        (0, )+clinmap_dup.generator_mapping())
    del clinmap_dup

    clintemp.optimize_rms(consonances.primes, consonances)
    rmserr = lintemp.getRMSError()
    crmserr = clintemp.rms_error(consonances.primes, consonances)
    assert abs(rmserr-crmserr) < 1e-10
    crmserr = clintemp.rms_error(
            [1.0] + consonances.primes,
            map(temper.reduceWithinTritone, consonances))
    assert abs(rmserr-crmserr) < 1e-10


def test_search(temper, limit, complexity, error, et_error):
    ets = temper.getLimitedETs(limit, 20, et_error*2)
    cets = get_limited_ets(limit, limit.primes, 20, et_error*2)
    # not only do the cets come out in reverse order,
    # but different temperaments with the same number of notes to the octave
    # are different as well, so a straight reverse won't work
    et_bases = []
    for et in ets:
        et_bases.append(tuple(et.basis))
    et_bases.sort()
    cet_bases = []
    for et in cets:
        cet_bases.append(tuple(et.mapping))
    cet_bases.sort()
    assert et_bases==cet_bases

    ets = temper.getLimitedETs(limit, 20, et_error)
    cets = get_limited_ets(limit, limit.primes, 20, et_error)
    et_bases = []
    for et in ets:
        et_bases.append(tuple(et.basis))
    et_bases.sort()
    cet_bases = []
    for et in cets:
        cet_bases.append(tuple(et.mapping))
    cet_bases.sort()
    assert et_bases==cet_bases

    def rms_fod(self, consonances=None):
        error = self.getRMSError(consonances)
        width = self.getComplexity(consonances)
        m, n = self.primesByMelody[0]
        return width*width*error + 1e-10*(1.618*m+n)

    lts = temper.getLinearTemperaments(ets, limit, rms_fod,
            complexity, error, 1)[:10]
    clts = get_linear_mappings(cets,
            map(temper.reduceWithinTritone, limit), limit.primes,
            complexity, error, 0, 9)
    lt_invariants = []
    for lt in lts:
        lt_invariants.append(lt.invariant())
    clt_invariants = []
    for clt in clts:
        clt_invariants.append(clt.invariant())
    assert (lt_invariants) # make sure we actually did something
    assert lt_invariants == clt_invariants


if __name__=="__main__":
    import temper
    primes = temper.limit5.primes

    test_examples(temper, primes)
    test_linear_examples(temper, primes, 12, 19, temper.limit5)
    test_linear_examples(temper, primes, 24, 19, temper.limit5)
    test_linear_examples(temper, primes, 29, 58, temper.limit15)
    test_search(temper, temper.limit11, 30, 0.01, 0.5)
    test_search(temper, temper.limit11, 100, 0.001, 0.3)
    test_search(temper, temper.limit5, 10, 0.05, 0.7)
    test_search(temper, temper.limit21, 100, 0.01, 0.6)
